r"""
The :py:mod:`vc2_conformance.constraint_table` module describes a constraint
system which is used to describe restrictions imposed by VC-2 levels. See
:py:data:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS` for the actual
level constraints table.

Tutorial
--------

Constraint tables enumerate allowed combinations of values as a list of
dictionaries containing :py:class:`ValueSet` objects. Each dictionary describes
a valid combination of values. In the contrived running example below we'll
define valid food-color combinations (rather than VC-2 codec options)::

    >>> real_foods = [
    ...     {"type": ValueSet("tomato"), "color": ValueSet("red")},
    ...     {"type": ValueSet("apple"), "color": ValueSet("red", "green")},
    ...     {"type": ValueSet("beetroot"), "color": ValueSet("purple")},
    ... ]

We can check dictionaries of values against this permitted list of
combinations using :py:func:`is_allowed_combination`::

    >>> # Allowed combinations
    >>> is_allowed_combination(real_foods, {"type": "tomato", "color": "red"})
    True
    >>> is_allowed_combination(real_foods, {"type": "apple", "color": "red"})
    True
    >>> is_allowed_combination(real_foods, {"type": "apple", "color": "green"})
    True

    >>> # Purple apples? I don't think so...
    >>> is_allowed_combination(real_foods, {"type": "apple", "color": "purple"})
    False

But we don't have to check a complete set of values. For example, we can check if a
particular color is valid for any foodstuff::

    >>> is_allowed_combination(real_foods, {"color": "red"})
    True
    >>> is_allowed_combination(real_foods, {"color": "yellow"})
    False

This behaviour allows us to detect the first non-constraint-satisfying value
when values are obtained sequentially (as they are for a VC-2 bitstream). The
bitstream validator (:py:mod:`vc2_conformance.decoder`) uses this functionality
to check bitstream values conform to the constraints imposed by a specified
VC-2 level.

Given an incomplete set of values, we can use :py:func:`allowed_values_for` to
discover what values are permissible for values we've not yet assigned. For
example:

    >>> # If we have an apple, what colors can it be?
    >>> allowed_values_for(real_foods, "color", {"type": "apple"})
    ValueSet('red', 'green')

    >>> # If we have something red, what might it be?
    >>> allowed_values_for(real_foods, "type", {"color": "red"})
    ValueSet('apple', 'tomato')

This functionality is used by the test case generators and encoder
(:py:mod:`vc2_conformance.test_cases` and :py:mod:`vc2_conformance.encoder`) to
discover combinations of bitstream features which satisfy particular level
requirements.


:py:class:`ValueSet`
--------------------

.. autoclass:: ValueSet
    :members:
    :special-members: __init__, __contains__, __eq__, __add__, __iter__, __str__

.. autoclass:: AnyValue


Constraint tables
-----------------

A 'constraint table' is defined as a list of 'allowed combination' dictionaries.

An 'allowed combination' dictionary defines a :py:class:`ValueSet` for every
permitted key.

For example, the following is a constraint table containing three allowed
combination dictionaries::

    >>> real_foods = [
    ...     {"type": ValueSet("tomato"), "color": ValueSet("red")},
    ...     {"type": ValueSet("apple"), "color": ValueSet("red", "green")},
    ...     {"type": ValueSet("beetroot"), "color": ValueSet("purple")},
    ... ]

A set of values satisfies a constraint table if there is at least one allowed
combination dictionary which contains the specified combination of values. For
example, the following two dictionaries satisfy the constraint table::

    {"type": "apple", "color": "red"}

    {"color": "red"}

The first satisfies the constraint table because the combination of values
given appears in the second entry of the constraint table.

The second satisfies the constraint table because, even though it does not
define a value for every key, the key it does define is included in both the
first and second entries.

Meanwhile, the following dictionaries do *not* satisfy the constraint table::

    {"type": "apple", "color": "purple"}

    {"type": "beetroot", "color": "purple", "pickleable": True}

The first of these contains values which, in isolation, would be permitted by
the second and third entries of the table but which are not present together in
any table entries. Consequently, this value does not satisfy the table.

The second contains a 'pickleable' key which is not present in any of the
allowed combinations in constraint table and so does not satisfy the table.

The functions below may be used to interrogate a constraint table.

.. autofunction:: filter_constraint_table

.. autofunction:: is_allowed_combination

.. autofunction:: allowed_values_for


CSV format
----------

A Constraint table can be read from CSV files using the following function:

.. autofunction:: read_constraints_from_csv

"""

import csv

from vc2_data_tables.csv_readers import open_utf8, is_ditto


__all__ = [
    "ValueSet",
    "AnyValue",
    "filter_constraint_table",
    "is_allowed_combination",
    "allowed_values_for",
    "read_constraints_from_csv",
]


class ValueSet(object):
    """
    Represents a set of allowed values. May consist of anything from a single
    value, a range of values or a combination of several of these.
    """

    def __init__(self, *values_and_ranges):
        """
        Create a :py:class:`ValueSet` containing the specified set of values::

            >>> no_values = ValueSet()
            >>> 100 in no_values
            False

            >>> single_value = ValueSet(100)
            >>> 100 in single_value
            True
            >>> 200 in single_value
            False

            >>> range_of_values = ValueSet((10, 20))
            >>> 9 in range_of_values
            False
            >>> 10 in range_of_values
            True
            >>> 11 in range_of_values
            True
            >>> 20 in range_of_values  # NB: Range is inclusive
            True
            >>> 21 in range_of_values
            False

            >>> many_values = ValueSet(100, 200, (300, 400))
            >>> 100 in many_values
            True
            >>> 200 in many_values
            True
            >>> 300 in many_values
            True
            >>> 350 in many_values
            True
            >>> 500 in many_values
            False

            >>> non_numeric = ValueSet("foo", "bar", "baz")
            >>> "foo" in non_numeric
            True
            >>> "nope" in non_numeric
            False

        Parameters
        ==========
        *values_and_ranges : value, or (lower_value, upper_value)
            Sets the initial set of values and (inclusive) ranges to be matched
        """
        # Individual values explicitly included in this value set
        self._values = set()

        # A series of (lower_bound, upper_bound) tuples which give *incuslive*
        # ranges of values which are permitted.
        self._ranges = set()

        for value_or_range in values_and_ranges:
            if isinstance(value_or_range, tuple):
                self.add_range(*value_or_range)
            else:
                self.add_value(value_or_range)

    def add_value(self, value):
        """
        Add a single value to the set.
        """
        # Don't add duplicates
        if value not in self:
            self._values.add(value)

    def add_range(self, lower_bound, upper_bound):
        """
        Add the range of values between the two inclusive bounds to the set.
        """
        # Remove any single values which are encompassed by this new range
        for value in list(self._values):
            if lower_bound <= value <= upper_bound:
                self._values.remove(value)

        # Combine this range with any existing ranges where possible
        ranges_to_remove = []
        for other_lower_bound, other_upper_bound in self._ranges:
            if lower_bound <= other_upper_bound and other_lower_bound <= upper_bound:
                ranges_to_remove.append((other_lower_bound, other_upper_bound))
                lower_bound = min(lower_bound, other_lower_bound)
                upper_bound = max(upper_bound, other_upper_bound)
        for lower_upper in ranges_to_remove:
            self._ranges.remove(lower_upper)

        self._ranges.add((lower_bound, upper_bound))

    def __contains__(self, value):
        """
        Test if a value is a member of this set. For example::

            >>> value_set = ValueSet(1, 2, 3)
            >>> 1 in value_set
            True
            >>> 100 in value_set
            False
        """
        if value in self._values:
            return True

        for lower_bound, upper_bound in self._ranges:
            if lower_bound <= value <= upper_bound:
                return True

        return False

    def is_disjoint(self, other):
        """
        Test if this :py:class:`ValueSet` is disjoint from another -- i.e. they
        share no common values.
        """
        if isinstance(other, AnyValue):
            # Special case: only the empty ValueSet is disjoint from the
            # AnyValue value set.
            if self._values or self._ranges:
                return False
            else:
                return True
        else:
            # Check for values which overlap
            for a, b in [(self, other), (other, self)]:
                for value in a._values:
                    if value in b:
                        return False

            # Check for ranges which overlap
            for a, b in [(self, other), (other, self)]:
                for start, end in a._ranges:
                    if start in b:
                        return False
                    if end in b:
                        return False

            return True

    def __hash__(self):
        # Warning: This null-hash function is provided only to allow ValueSets
        # to be used as part of dictionary keys in non-performance sensitive
        # settings.
        return hash(0)

    def __eq__(self, other):
        return (
            not isinstance(other, AnyValue)
            and self._values == other._values
            and self._ranges == other._ranges
        )

    def __ne__(self, other):
        """Required under Python 2.x"""
        return not self.__eq__(other)

    def __add__(self, other):
        """
        Combine two :py:class:`ValueSet` objects into a single object
        containing the union of both of their values.

        For example::

            >>> a = ValueSet(123)
            >>> b = ValueSet((10, 20))
            >>> a + b
            ValueSet(123, (10, 20))
        """
        if isinstance(other, AnyValue):
            return AnyValue()
        else:
            out = type(self)()

            for value in self._values:
                out.add_value(value)
            for value in other._values:
                out.add_value(value)

            for lower, upper in self._ranges:
                out.add_range(lower, upper)
            for lower, upper in other._ranges:
                out.add_range(lower, upper)

            return out

    def __iter__(self):
        """
        Iterate over the values and (lower_bound, upper_bound) tuples in this
        value set in no particular order.
        """
        for value in self._values:
            yield value
        for range in self._ranges:
            yield range

    def iter_values(self):
        """
        Iterate over the values (including the enumerated values of ranges) in
        this value set in no particular order.
        """
        for value in self._values:
            yield value
        for low, high in self._ranges:
            for value in range(low, high + 1):
                yield value

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            ", ".join(map(repr, self)),
        )

    def __str__(self):
        """
        Produce a human-readable description of the permitted values.

        For example::

            >>> print(ValueSet())
            {<no values>}
            >>> print(ValueSet(1, 2, 3, (10, 20)))
            {1, 2, 3, 10-20}
        """
        values_and_ranges = sorted([(v,) for v in self._values] + list(self._ranges))
        if len(values_and_ranges) == 0:
            return "{<no values>}"
        else:
            return "{{{}}}".format(
                ", ".join(
                    repr(value[0]) if len(value) == 1 else "{!r}-{!r}".format(*value)
                    for value in values_and_ranges
                )
            )


class AnyValue(ValueSet):
    """
    Like :py:class:`ValueSet` but represents a 'wildcard' set of values which
    contains all possible values.
    """

    def __init__(self):
        pass

    def add_value(self, value):
        pass

    def add_range(self, lower_bound, upper_bound):
        pass

    def __contains__(self, value):
        return True

    def is_disjoint(self, other):
        if isinstance(other, AnyValue):
            return False
        else:
            return other.is_disjoint(self)

    def __hash__(self):
        return hash(0)

    def __eq__(self, other):
        return isinstance(other, AnyValue)

    def __add__(self, other):
        return AnyValue()

    def __iter__(self):
        raise AttributeError("__iter__")

    def iter_values(self):
        raise AttributeError("iter_values")

    def __repr__(self):
        return "{}()".format(type(self).__name__)

    def __str__(self):
        return "{<any value>}"


def filter_constraint_table(constraint_table, values):
    """
    Return the subset of ``constraint_table`` entries which match all of the
    values in ``values``. That is, with the entries whose constraints are not
    met by the provided values removed.
    """
    return [
        allowed_combination
        for allowed_combination in constraint_table
        if all(
            key in allowed_combination and value in allowed_combination[key]
            for key, value in values.items()
        )
        or len(allowed_combination) == 0  # Special case: 'catch all' rule
    ]


def is_allowed_combination(constraint_table, values):
    """
    Check to see if the ``values`` dictionary holds an allowed combination of
    values according to the provided constraint table.

    .. note::

        A candidate containing only a subset of the keys listed in the
        constraint table is allowed if the fields it does define are a
        permitted combination.

    Parameters
    ==========
    constraint_table: [{key: :py:class:`ValueSet`, ...}, ...]
    values : {key: value}
    """
    return len(filter_constraint_table(constraint_table, values)) > 0


def allowed_values_for(constraint_table, key, values={}, any_value=AnyValue()):
    """
    Return a :py:class:`ValueSet` which matches all allowed values for the
    specified key, given the existing values defined in ``values``.

    Parameters
    ==========
    constraint_table : [{key: :py:class:`ValueSet`, ...}, ...]
    key : key
    values : {key: value, ...}
        Optional. The values already chosen. (Default: assume nothing chosen).
    any_value : :py:class:`ValueSet`
        Optional. If provided and :py:class:`AnyValue` is allowed, this value
        will be substituted instead. This may be useful when
        :py:class:`AnyValue` is being used as a short-hand for a more concrete
        set of values.
    """
    out = ValueSet()

    for allowed_combination in filter_constraint_table(constraint_table, values):
        out += allowed_combination.get(key, ValueSet())

    if isinstance(out, AnyValue):
        return any_value
    else:
        return out


def read_constraints_from_csv(csv_filename):
    r'''
    Reads a table of constraints from a CSV file.

    The CSV file should be arranged with each row describing a particular value
    to be constrained and each column defining an allowed combination of
    values.

    Empty rows and rows containing only '#' prefixed values will be skipped.

    The first column will be treated as the keys being constrained, remaining
    columns should contain allowed combinations of values. Each of these values
    will be converted into a :py:class:`ValueSet` as follows:

    * Values which contain integers will be converted to ``int``
    * Values which contain 'TRUE' or 'FALSE' will be converted to ``bool``
    * Values containing a pair of integers separated by a ``-`` will be treated
      as an incusive range.
    * Several comma-separated instances of the above will be combined into a
      single :py:class:`ValueSet`. (Cells containing comma-separated values
      will need to be enclosed in double quotes (``"``) in the CSV).
    * The value 'any' will be substituted for :py:class:`AnyValue`.
    * Empty cells will be converted into empty :py:class:`ValueSets <ValueSet>`.
    * Cells which contain only a pair of quotes (e.g. ``"``, i.e. ditto) will
      be assigned the same value as the column to their left. (This is encoded
      using four double quotes (``""""``) in CSV format).

    The read constraint table will be returned as a list of dictionaries (one
    per column) as expected by the functions in
    :py:mod:`vc2_conformance.constraint_table`.
    '''
    out = []

    with open_utf8(csv_filename) as f:
        for row in csv.reader(f):
            # Skip empty lines
            if all(not cell.strip() or cell.strip().startswith("#") for cell in row):
                continue

            # Add extra constraint sets as required
            for _ in range(len(out), len(row) - 1):
                out.append({})

            # Populate this row's values
            key = row[0]
            last_value = ValueSet()
            for i, column in enumerate(row[1:]):
                value = ValueSet()
                if is_ditto(column):
                    value += last_value
                elif column.strip().lower() == "any":
                    value = AnyValue()
                else:
                    for value_string in column.split(","):
                        values = [
                            True
                            if s.strip().lower() == "true"
                            else False
                            if s.strip().lower() == "false"
                            else int(s)
                            for s in value_string.partition("-")[::2]
                            if s
                        ]
                        if len(values) == 1:
                            value.add_value(values[0])
                        elif len(values) == 2:
                            value.add_range(values[0], values[1])
                out[i][key] = value
                last_value = value

    return out



"""
The :py:mod:`vc2_conformance.decoder.assertions` module contains functions
which perform checks which are used in multiple places within a bitstream and
have been factored-out to avoid repetition.
"""

from collections import OrderedDict

from vc2_data_tables import (
    ParseCodes,
    PictureCodingModes,
)

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.version_constraints import MINIMUM_MAJOR_VERSION

from vc2_conformance.symbol_re import WILDCARD, END_OF_SEQUENCE

from vc2_conformance.constraint_table import allowed_values_for

from vc2_conformance.decoder.exceptions import (
    ValueNotAllowedInLevel,
    NonConsecutivePictureNumbers,
    EarliestFieldHasOddPictureNumber,
    MajorVersionTooHigh,
)


def assert_in(value, collection, exception_type, *args):
    """
    Check to see if a value is a member of the provided collection using
    ``in``. If it is not, throws an exception of the given type with the value
    and collection passed as as arguments.
    """
    if value not in collection:
        raise exception_type(value, collection, *args)


def assert_in_enum(value, enum, exception_type):
    """
    Check to see if a value is a member of the provided assertion type. If it
    is not, throws an exception of the given type with the value as argument.
    """
    try:
        enum(value)
    except ValueError:
        raise exception_type(value)


def assert_parse_code_in_sequence(parse_code, matcher, exception_type, *args):
    """
    Check that the specified parse code's name matches the next value in the
    sequence defined by :py:class:`vc2_conformance.symbol_re.Matcher`.

    If it does not, ``exception_type`` will be raised with the following
    arguments:

    * The ``parse_code`` (as a :py:class:`~vc2_data_tables.ParseCodes`)
    * A list of parse codes (as a
      :py:class:`~vc2_data_tables.ParseCodes`) which would have been
      valid (or ``None`` if any parse code would be allowed.
    * A boolean which is True if it would have been valid to end the sequence
      at this point.
    * Any additional arguments passed to this function
    """
    parse_code = ParseCodes(parse_code)

    if not matcher.match_symbol(parse_code.name):
        expected_parse_codes = []
        expected_end = False
        for parse_code_name in matcher.valid_next_symbols():
            if parse_code_name == WILDCARD:
                expected_parse_codes = None
                break
            elif parse_code_name == END_OF_SEQUENCE:
                expected_end = True
            elif expected_parse_codes is not None:
                expected_parse_codes.append(getattr(ParseCodes, parse_code_name))

        raise exception_type(parse_code, expected_parse_codes, expected_end, *args)


def assert_parse_code_sequence_ended(matcher, exception_type, *args):
    """
    Check that the specified :py:class:`vc2_conformance.symbol_re.Matcher` has
    reached a valid end for the sequence of parse codes specified.

    If the matcher still expects more parse codes, ``exception_type`` will be
    raised with the following arguments:

    * The ``parse_code`` will be None
    * A list of parse codes (as a
      :py:class:`~vc2_data_tables.ParseCodes`) which would have been
      valid (or ``None`` if any parse code would be allowed.
    * A boolean which is True if it would have been valid to end the sequence
      at this point. (Always False, in this case)
    * Any additional arguments passed to this function
    """
    if not matcher.is_complete():
        expected_parse_codes = []
        for parse_code_name in matcher.valid_next_symbols():
            if parse_code_name == WILDCARD:
                expected_parse_codes = None
                break
            elif expected_parse_codes is not None:
                expected_parse_codes.append(getattr(ParseCodes, parse_code_name))

        raise exception_type(None, expected_parse_codes, False, *args)


def assert_level_constraint(state, key, value):
    """
    Check that the given key and value is allowed according to the
    :py:data:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS`. Throws a
    :py;exc:`vc2_conformance.decoder.exceptions.ValueNotAllowedInLevel`
    exception on failure.

    Takes the current :py:class:`~vc2_conformance.pseudocode.state.State` instance from
    which the current
    :py:attr:`~vc2_conformance.pseudocode.state.State._level_constrained_values` will be
    created/updated.
    """
    state.setdefault("_level_constrained_values", OrderedDict())

    allowed_values = allowed_values_for(
        LEVEL_CONSTRAINTS,
        key,
        state["_level_constrained_values"],
    )

    if value not in allowed_values:
        raise ValueNotAllowedInLevel(
            state["_level_constrained_values"], key, value, allowed_values
        )
    else:
        state["_level_constrained_values"][key] = value


def assert_picture_number_incremented_as_expected(state, picture_number_offset):
    """
    Check that ``state["picture_number"]`` has been set to an appropriate value
    by (12.2) picture_header or (14.2) fragment_header.  It should be called
    after the picture number has been read from the stream into
    ``state["picture_number"]`` and with the
    :py:func:`~vc2_conformance.decoder.io.tell` value just after the value was
    read.

    This assertion will check that:

    * (12.2), (14.2) Picture numbers in a sequence must increment by 1 and wrap after
      2^32-1 back to zero, throwing :py:exc:`NonConsecutivePictureNumbers` if
      this is not the case.
    * (12.2), (14.2) When coded as fields, the first field in the sequence must
      have an even picture number, throwing
      :py:exc:`EarliestFieldHasOddPictureNumber` if this is not the case.

    This assertion also has the following side-effects:

    * Sets ``state["_last_picture_number"]`` to ``state["picture_number"]``
    * Sets ``state["_last_picture_number_offset"]`` to the value passed in the
      picture_number_offset argument.
    * Increments ``state["_num_pictures_in_sequence"]``

    In the case of fragments (14.2), this assertion should be called only for
    the first fragment in a picture (with fragment_slice_count==0).
    """
    # (12.2), (14.4) Picture numbers in a sequence must increment by 1 and wrap
    # after 2^32-1 back to zero
    if "_last_picture_number" in state:
        expected_picture_number = (state["_last_picture_number"] + 1) & 0xFFFFFFFF
        if state["picture_number"] != expected_picture_number:
            raise NonConsecutivePictureNumbers(
                state["_last_picture_number_offset"],
                state["_last_picture_number"],
                picture_number_offset,
                state["picture_number"],
            )
    state["_last_picture_number"] = state["picture_number"]
    state["_last_picture_number_offset"] = picture_number_offset

    # (12.2), (14.4) When coded as fields, the first field in the sequence must
    # have an even picture number.
    if state["picture_coding_mode"] == PictureCodingModes.pictures_are_fields:
        early_field = state["_num_pictures_in_sequence"] % 2 == 0
        even_number = state["picture_number"] % 2 == 0
        if early_field and not even_number:
            raise EarliestFieldHasOddPictureNumber(state["picture_number"])
    state["_num_pictures_in_sequence"] += 1


def log_version_lower_bound(state, minimum_major_version):
    """
    Records an instance where a value in the stream requires that
    major_version have a certain value. These will later be checked by
    :py:func:`assert_major_version_is_minimal`. Bounds are logged in
    ``state["_expected_major_version"]``

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
    minimum_major_version : int
        The minimum major version number allowed to support the feature being
        logged.
    """
    state["_expected_major_version"] = max(
        state.get("_expected_major_version", MINIMUM_MAJOR_VERSION),
        minimum_major_version,
    )


def assert_major_version_is_minimal(state):
    """
    Checks whether the major_version number supplied is the lowest possible
    major_version number for this stream (as required by (11.2.2)). Should be
    called once at the end of a stream.
    """
    major_version = state["major_version"]
    expected_major_version = state.get("_expected_major_version", MINIMUM_MAJOR_VERSION)

    # Errata: The following special case was missing from the spec
    #
    # (11.2.2) Special case for empty sequences of fragmented pictures: When an
    # encoder is configured to produce fragmented pictures in streams the
    # major_version needs to be set to '3' to allow fragmented picture data
    # units. In the case of an empty sequence, major_version may not have to be
    # '3' since no data units with a fragment parse code will be present. To
    # make the life of encoders developers easier in this special case, the
    # spec permits them to set major_version to '3' even when no fragmented
    # picture data units appear in the stream enabling them to use the same
    # sequence header for all sequences, regardless of length.
    if state["_num_pictures_in_sequence"] == 0 and major_version == 3:
        return

    if major_version > expected_major_version:
        raise MajorVersionTooHigh(major_version, expected_major_version)



"""
The main VC-2 conformance software is contained within the
:py:mod:`vc2_conformance` module.

..
    You are currently reading VC-2 conformance documentation in its source form
    (e.g. directly from the Python source docstrings or via ``help()``). You
    may prefer to read the HTML/PDF version of this documentation where images
    and cross references will be displayed more usefully.

.. note::

    Before going any further you should familiarize yourself with the
    :ref:`user-guide` which gives an introduction of the general tasks carried
    out by the conformance software.

Below we give a general overview of the design of the conformance software.


Main components
---------------

The VC-2 conformance software consists of four main components:

* A reference VC-2 decoder, including bitstream validation logic
  (:py:mod:`vc2_conformance.decoder`)
* A flexible VC-2 encoder (:py:mod:`vc2_conformance.encoder`)
* A VC-2 bitstream manipulation library (:py:mod:`vc2_conformance.bitstream`)
* A set of test case generation routines (:py:mod:`vc2_conformance.test_cases`)

The reference decoder and bitstream validator forms a key part of the
conformance testing procedures (via the :ref:`vc2-bitstream-validator`
command). This decoder is based directly on the pseudocode published in the
VC-2 standard. Refer to the :py:mod:`vc2_conformance.decoder` module
documentation for a full introduction to this component.

The encoder is used internally to generate decoder test cases. This encoder is
flexible enough to support all of VC-2 features but otherwise simplistic in
terms of runtime performance and picture quality. Further documentation on the
encoder's design and functionality can be found in the
:py:mod:`vc2_conformance.encoder` module.

The bitstream manipulation library in the :py:mod:`vc2_conformance.bitstream`
module is used internally for three purposes.  Firstly it is used during
decoder test case generation to produce bitstreams with specific properties.
Secondly it is used by the :ref:`vc2-bitstream-viewer` command to produce human
readable descriptions of bitstream contents. Finally it is used extensively by
the VC-2 conformance software's own test suite to generate and check
bitstreams.

Finally, the test case generation routines form the basis of the
:ref:`vc2-test-case-generator` tool which generates test pictures and
bitstreams used during conformance testing procedures. These will be introduced
in :ref:`maintainer-test-case-generator`.


Use of VC-2 pseudocode
----------------------

The VC-2 conformance software design prioritises correctness and consistency
with the VC-2 specification. To achieve this, significant parts are built using
the pseudocode within the VC-2 specification.

The VC-2 specification uses pseudocode to define the nominal operation of a
VC-2 decoder. The pseudocode language used is sufficiently similar to Python
that a translation from pseudocode into executable Python is trivial. Automated
translation is also possible using the `VC-2 Pseudocode Parser tool
<https://github.com/bbc/vc2_pseudocode_parser>`_. Once translated, the
pseudocode may be used as the basis for correct-by-definition implementations
of parts of a VC-2 codec.

The reference VC-2 decoder and bitstream validator
(:py:mod:`vc2_conformance.decoder`) consists of the VC-2 pseudocode
(implementing the decoder behaviour) augmented with additional checks
(implementing the validation logic).

The VC-2 encoder (:py:mod:`vc2_conformance.encoder`), though not specified by
the VC-2 standard, nevertheless makes substantial use of the VC-2 pseudocode
functions. For example routines for computing slice dimensions are reused while
other parts, such as the forward discrete wavelet transform are simple
inversions of the decoder pseudocode. The correctness of these inversions is
relatively easily verified in the test suite thanks to the invertability of
VC-2's transforms.

The bitstream manipulation library (:py:mod:`vc2_conformance.bitstream`)
provides routines for serialising and deserialising binary bitstreams into
easily manipulated Python data structures.  With the exception of the Python
data structure definitions, this library is based entirely on the VC-2
pseudocode, using the :py:mod:`~vc2_conformance.bitstream.serdes` framework.

To ensure consistency with the VC-2 pseudocode, the conformance software's test
suite automatically compares the software's source code with the pseudocode to
verify equivalence. See the :py:mod:`verification` module for details.


Performance
-----------

The major drawback of the pseudocode-driven approach used by this software is
poor performance. The VC-2 pseudocode is structured with comprehensibility as
its main priority and therefore often has poor algorithmic performance.
Further, the use of Python -- due to its similarity with the pseudocode and
high level of abstraction -- introduces an additional performance overhead.

Another source of slow performance is the use of infinite precision (i.e.
native Python) integers whenever possible. This helps avoid the class of bugs
relating to the use of insufficient integer bit width in most calculations. The
dynamic range of signals passing through a VC-2 codec can grow dramatically
(i.e. by many orders of magnitude) and vary significantly between superficially
similar inputs. As a consequence, this class of bug is all too easy to
introduce without infinite precision arithmetic.

As a result of the above factors, the conformance software may take on the
order minutes to encode or decode each picture in a stream. While this would be
unacceptable for a production encoder or decoder, the conformance software is
intended only for use with only extremely short sequences, where correctness is
the most significant factor.

The majority of test cases apply to decoders for which all necessary materials
(test bitstreams and reference decodings) may be produced in an 'overnight'
batch process after which the tools are not needed.  The remaining (encoder)
test cases generally amount to fewer than ten frames.


Test case generation
--------------------

Since VC-2 supports a great variety of configurations, parametrised test case
generators (:py:mod:`vc2_conformance.test_cases`) are used rather than a
'universal' collection of bitstreams. While this means that users of the
conformance software are required to configure and run the test case generator
themselves, it also simplifies the testing process by ensuring only relevant
test cases are produced. Furthermore, it allows certain tests to be highly
tailored to a particular configuration.  For example, signal range tests are
targeted specifically at the specific wavelet transform configuration, bit
width and quantization matrices used.


External packages
-----------------

Some smaller, or more specialised aspects of the conformance testing software
have been split into their own separate Python packages. These are:

:py:mod:`vc2_conformance_data`
    Larger data files (including pictures and precomputed values) which are
    used to generate certain test cases.

:py:mod:`vc2_data_tables`
    General VC-2 related constants (e.g.
    :py:data:`~vc2_data_tables.PARSE_INFO_PREFIX`), enumerated values (e.g.
    :py:class:`~vc2_data_tables.ParseCodes`) and tabular data used during
    coding (e.g. :py:data:`~vc2_data_tables.PRESET_FRAME_RATES`).

:py:mod:`vc2_bit_widths`
    Computes near worst-case inputs for VC-2 encoders and decoders which
    produce very large signal values. This package is used to generate test
    cases to verify codecs have used large enough integers in their
    implementations.




"""

from vc2_conformance.version import __version__



"""
Quantization-related VC-2 pseudocode routines (13.3).
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.pseudocode.vc2_math import sign

__all__ = [
    "inverse_quant",
    "forward_quant",
    "quant_factor",
    "quant_offset",
]


@ref_pseudocode
def inverse_quant(quantized_coeff, quant_index):
    """(13.3.1)"""
    magnitude = abs(quantized_coeff)
    if magnitude != 0:
        magnitude *= quant_factor(quant_index)
        magnitude += quant_offset(quant_index)
        magnitude += 2
        magnitude //= 4
    return sign(quantized_coeff) * magnitude


@ref_pseudocode(deviation="inferred_implementation")
def forward_quant(coeff, quant_index):
    """(13.3.1) Based on the informative note 1."""
    magnitude = abs(coeff)
    if coeff >= 0:
        return (4 * magnitude) // quant_factor(quant_index)
    else:
        return -((4 * magnitude) // quant_factor(quant_index))


@ref_pseudocode
def quant_factor(index):
    """(13.3.2)"""
    base = 2 ** (index // 4)
    if (index % 4) == 0:
        return 4 * base
    elif (index % 4) == 1:
        return ((503829 * base) + 52958) // 105917
    elif (index % 4) == 2:
        return ((665857 * base) + 58854) // 117708
    elif (index % 4) == 3:
        return ((440253 * base) + 32722) // 65444


@ref_pseudocode
def quant_offset(index):
    """(13.3.2)"""
    if index == 0:
        offset = 1
    elif index == 1:
        offset = 2
    else:
        offset = (quant_factor(index) + 1) // 2
    return offset



"""
The :py:mod:`vc2_conformance.decoder.exceptions` module defines a number of
exceptions derived from :py:exc:`ConformanceError` representing different
conformance errors a bitstream may contain. These exceptions provide additional
methods which return detailed human-readable information about the conformance
error.

.. autoexception:: ConformanceError
    :members:

"""

from textwrap import dedent

from fractions import Fraction

from vc2_conformance.string_utils import wrap_paragraphs

from vc2_conformance.pseudocode.vc2_math import intlog2

from vc2_data_tables import (
    PARSE_INFO_PREFIX,
    PARSE_INFO_HEADER_BYTES,
    PRESET_FRAME_RATES,
    ParseCodes,
    Profiles,
    LEVELS,
    PresetSignalRanges,
    PresetColorSpecs,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
    WaveletFilters,
)

from vc2_conformance.level_constraints import LEVEL_SEQUENCE_RESTRICTIONS

from vc2_conformance.version_constraints import (
    MINIMUM_MAJOR_VERSION,
    preset_frame_rate_version_implication,
    preset_signal_range_version_implication,
    preset_color_spec_version_implication,
    preset_color_primaries_version_implication,
    preset_color_matrix_version_implication,
    preset_transfer_function_version_implication,
    parse_code_version_implication,
    profile_version_implication,
)

from vc2_conformance.bitstream.io import to_bit_offset


def known_parse_code_to_string(parse_code):
    """
    Convert a known parse code (i.e. one in
    :py:class:`~vc2_data_tables.ParseCodes`) into a string such as::

        "end of sequence (0x10)"
    """
    return "{} (0x{:02X})".format(
        ParseCodes(parse_code).name.replace("_", " "),
        parse_code,
    )


def known_profile_to_string(profile):
    """
    Convert a known profile number (i.e. one in
    :py:class:`~vc2_data_tables.Profiles`) into a string such as::

        "high quality (3)"
    """
    return "{} ({:d})".format(
        Profiles(profile).name.replace("_", " "),
        profile,
    )


def known_wavelet_to_string(wavelet_index):
    """
    Convert a known wavelet index (i.e. one in
    :py:class:`~vc2_data_tables.WaveletFilters`) into a string such as::

        "Haar With Shift (4)"
    """
    return "{} ({:d})".format(
        WaveletFilters(wavelet_index).name.replace("_", " ").title(),
        wavelet_index,
    )


def explain_parse_code_sequence_structure_restrictions(
    actual_parse_code,
    expected_parse_codes,
    expected_end,
):
    """
    Produce an sentence explaining how a particular sequence requirement has
    been violated.

    Parameters
    ==========
    actual_parse_code : None, :py:class:`vc2_data_tables.ParseCodes`
        The (valid) parse code which was encountered which violated the stated
        sequencing requirements. If None, a premature end of sequence was encountered.
    expected_parse_codes : None or [:py:class:`vc2_data_tables.ParseCodes`, ...]
        If a list, enumerates the (valid) parse codes which would have been
        allowed instead. If None, any parse code would have been allowed.
    expected_end : bool
        If True, indicates that the sequence was expected to end (i.e. that no
        more parse codes were expected). If False, see expected_parse_codes.
    """
    return "{} was encountered but {} expected.".format(
        (
            "The parse code {}".format(
                known_parse_code_to_string(actual_parse_code),
            )
            if actual_parse_code is not None
            else "No further parse code"
        ),
        (
            "no further parse code"
            if expected_end is True
            else "any parse code"
            if expected_parse_codes is None
            else " or ".join(map(known_parse_code_to_string, expected_parse_codes))
        ),
    )


class ConformanceError(Exception):
    """
    Base class for all bitstream conformance failure exceptions.
    """

    def __str__(self):
        return wrap_paragraphs(self.explain()).partition("\n")[0]

    def explain(self):
        """
        Produce a detailed human readable explanation of the conformance
        failure.

        Should return a string which can be re-linewrapped by
        :py:func:`vc2_conformance.string_utils.wrap_paragraphs`.

        The first line will be used as a summary when the exception is printed
        using :py:func:`str`.
        """
        raise NotImplementedError()

    def bitstream_viewer_hint(self):
        """
        Return a set of sample command line arguments for the
        vc2-bitstream-viewer tool which will display the relevant portion of
        the bitstream.

        This string may include the following :py:meth:`str.format`
        substitutions which should be filled in before display:

        * ``{cmd}`` The command name of the bitstream viewer (i.e. usually
          ``vc2-bitstream-viewer``)
        * ``{file}`` The filename of the bitstream.
        * ``{offset}`` The bit offset of the next bit in the bitstream to be
          read.

        This returned string should *not* be line-wrapped but should be
        de-indented by :py:func:`textwrap.dedent`.
        """
        return """
            To view the offending part of the bitstream:

                {cmd} {file} --offset {offset}
        """

    def offending_offset(self):
        """
        If known, return the bit-offset of the offending part of the bitstream.
        Otherwise return None (and the current offset will be assumed).
        """
        return None


################################################################################
# Conformance failure exceptions
################################################################################


class UnexpectedEndOfStream(ConformanceError):
    """
    Reached the end of the stream while attempting to perform read operation.
    """

    def explain(self):
        return """
            Unexpectedly encountered the end of the stream.

            A VC-2 Stream shall be a concatenation of one or more VC-2
            sequences (10.3). Sequences shall end with a parse info header with
            an end of sequence parse code (0x10) (10.4.1)

            Did the sequence omit a terminating parse info with the end of
            sequence (0x10) parse code?
        """


class BadParseCode(ConformanceError):
    """
    parse_info (10.5.1) has been given an unrecognised parse code.

    The exception argument will contain the received parse code.
    """

    def __init__(self, parse_code):
        self.parse_code = parse_code
        super(BadParseCode, self).__init__()

    def explain(self):
        return """
            An invalid parse code, 0x{:02X}, was provided to a parse info
            header (10.5.1).

            See (Table 10.1) for the list of allowed parse codes.

            Perhaps this bitstream conforms to an earlier or later version of the
            VC-2 standard?
        """.format(
            self.parse_code
        )


class BadParseInfoPrefix(ConformanceError):
    """
    This exception is thrown when the parse_info (10.5.1) prefix value read
    from the bitstream doesn't match the expected value:

        The parse info prefix shall be 0x42 0x42 0x43 0x44

    The exception argument will contain the read prefix value.
    """

    def __init__(self, parse_info_prefix):
        self.parse_info_prefix = parse_info_prefix
        super(BadParseInfoPrefix, self).__init__()

    def explain(self):
        return """
            An invalid prefix, 0x{:08X}, was encountered in a parse info block
            (10.5.1). The expected prefix is 0x{:08X}.

            Is the parse_info block byte aligned (10.5.1)?

            Did the preceeding data unit over- or under-run the expected
            length? For example, were any unused bits in a picture slice filled
            with the correct number of padding bits (A.4.2)?
        """.format(
            self.parse_info_prefix,
            PARSE_INFO_PREFIX,
        )


class InconsistentNextParseOffset(ConformanceError):
    """
    This exception is thrown when a ``next_parse_offset`` value encoded in a
    parse_info (10.5.1) block does not match the offset of the next parse_info
    in the stream.

    Parameters
    ==========
    parse_info_offset : int
        The bitstream byte offset of the start of the parse_info block
        containing the bad next_parse_offset value.
    next_parse_offset : int
        The offending next_parse_offset value
    true_parse_offset : int
        The actual offset from parse_info_offset of the next parse_info block
        in the stream.
    """

    def __init__(self, parse_info_offset, next_parse_offset, true_parse_offset):
        self.parse_info_offset = parse_info_offset
        self.next_parse_offset = next_parse_offset
        self.true_parse_offset = true_parse_offset
        super(InconsistentNextParseOffset, self).__init__()

    def explain(self):
        return """
            Incorrect next_parse_offset value in parse info: got {} bytes,
            expected {} bytes (10.5.1).

            The erroneous parse info block begins at bit offset {} and is
            followed by the next parse info block at bit offset {}.

            Does the next_parse_offset include the {} bytes of the parse info
            header?

            Is next_parse_offset given in bits, not bytes?
        """.format(
            self.next_parse_offset,
            self.true_parse_offset,
            to_bit_offset(self.parse_info_offset),
            to_bit_offset(self.parse_info_offset + self.true_parse_offset),
            PARSE_INFO_HEADER_BYTES,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the erroneous parse info block:

                {{cmd}} {{file}} --offset {} --after-context 144

            To view the following parse info block:

                {{cmd}} {{file}} --offset {{offset}} --after-context 144
        """.format(
            to_bit_offset(self.parse_info_offset)
        )

    def offending_offset(self):
        return to_bit_offset(self.parse_info_offset)


class MissingNextParseOffset(ConformanceError):
    """
    This exception is thrown when a ``next_parse_offset`` value is given as
    zero but is not optional and must be provided.

    The parse_code for the offending parse info block is provided as an
    argument.
    """

    def __init__(self, parse_code):
        self.parse_code = parse_code
        super(MissingNextParseOffset, self).__init__()

    def explain(self):
        return """
            A next_parse_offset value of zero was provided in a {} (parse_code
            = 0x{:02X}) parse info block where a valid next_parse_offset value
            is mandatory (10.5.1).

            Does the next_parse_offset include the {} bytes of the parse info
            header?
        """.format(
            ParseCodes(self.parse_code).name.replace("_", " "),
            self.parse_code,
            PARSE_INFO_HEADER_BYTES,
        )


class InvalidNextParseOffset(ConformanceError):
    """
    This exception is thrown when a ``next_parse_offset`` value contains a
    value between 1 and 12 (inclusive). All of these byte offsets refer to an
    offset in the stream which is still within the current parse_info block.

    The exception argument will contain the offending next_parse_offset value.
    """

    def __init__(self, next_parse_offset):
        self.next_parse_offset = next_parse_offset
        super(InvalidNextParseOffset, self).__init__()

    def explain(self):
        return """
            Invalid next_parse_offset value {} found in parse info header
            (10.5.1).

            The next_parse_offset value must account for the {} bytes taken by
            the parse info block. As a consequence this value must be strictly
            {} or greater (except in circumstances where it may be omitted when
            it must be 0) (15.5.1).

            Does the next_parse_offset include the {} bytes of the parse info
            header?
        """.format(
            self.next_parse_offset,
            PARSE_INFO_HEADER_BYTES,
            PARSE_INFO_HEADER_BYTES,
            PARSE_INFO_HEADER_BYTES,
        )


class NonZeroNextParseOffsetAtEndOfSequence(ConformanceError):
    """
    This exception is thrown when an end-of-sequence defining parse_info
    (10.5.1) has a non-zero next_parse_offset.

    The exception argument will contain the offending next_parse_offset value.
    """

    def __init__(self, next_parse_offset):
        self.next_parse_offset = next_parse_offset
        super(NonZeroNextParseOffsetAtEndOfSequence, self).__init__()

    def explain(self):
        return """
            Non-zero next_parse_offset value, {}, in the parse info at the end
            of the sequence (10.5.1).

            Does the next_parse_offset incorrectly include an offset into an
            adjacent sequence?
        """.format(
            self.next_parse_offset,
        )


class InconsistentPreviousParseOffset(ConformanceError):
    """
    This exception is thrown when the ``previous_parse_offset`` value encoded
    in a parse_info (10.5.1) block does not match the offset of the previous
    parse_info in the stream.

    Parameters
    ==========
    last_parse_info_offset : int
        The bitstream byte offset of the start of the previous parse_info
        block.
    previous_parse_offset : int
        The offending previous_parse_offset value
    true_parse_offset : int
        The actual byte offset from the last parse_info to the current one in
        the stream.
    """

    def __init__(
        self, last_parse_info_offset, previous_parse_offset, true_parse_offset
    ):
        self.last_parse_info_offset = last_parse_info_offset
        self.previous_parse_offset = previous_parse_offset
        self.true_parse_offset = true_parse_offset
        super(InconsistentPreviousParseOffset, self).__init__()

    def explain(self):
        return """
            Incorrect previous_parse_offset value in parse info: got {} bytes,
            expected {} bytes (10.5.1).

            The erroneous parse info block begins at offset {} bits and follows
            a parse info block at offset {} bits.

            Does the previous_parse_offset include the {} bytes of the parse info
            header?

            Is previous_parse_offset given in bits, not bytes?

            Was the previous_parse_offset incorrectly omitted after a data unit
            whose size was not initially known?

            Was this parse info block copied from another sequence without
            updating the previous_parse_offset?
        """.format(
            self.previous_parse_offset,
            self.true_parse_offset,
            to_bit_offset(self.last_parse_info_offset + self.true_parse_offset),
            to_bit_offset(self.last_parse_info_offset),
            PARSE_INFO_HEADER_BYTES,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the erroneous parse info block:

                {{cmd}} {{file}} --offset {{offset}} --after-context 144

            To view the proceeding parse info block:

                {{cmd}} {{file}} --offset {} --after-context 144
        """.format(
            to_bit_offset(self.last_parse_info_offset)
        )


class NonZeroPreviousParseOffsetAtStartOfSequence(ConformanceError):
    """
    This exception is thrown when the first parse_info (10.5.1) has a non-zero
    previous_parse_offset.

    The exception argument will contain the offending previous_parse_offset value.
    """

    def __init__(self, previous_parse_offset):
        self.previous_parse_offset = previous_parse_offset
        super(NonZeroPreviousParseOffsetAtStartOfSequence, self).__init__()

    def explain(self):
        return """
            Non-zero previous_parse_offset, {}, in the parse info at the start
            of a sequence (10.5.1).

            Was this parse info block copied from another stream without the
            previous_parse_offset being updated?

            Does this parse info block incorrectly include an offset into an
            adjacent sequence?
        """.format(
            self.previous_parse_offset,
        )


class SequenceHeaderChangedMidSequence(ConformanceError):
    """
    This exception is thrown when a sequence_header (11.1) appears in the
    stream which does not match the previous sequence header byte-for-byte.

    Parameters
    ==========
    last_sequence_header_offset : int
    last_sequence_header_bytes : :py:class:`bytearray`
        The bitstream byte-offset and raw bytes for the previous
        sequence_header in the stream.
    this_sequence_header_offset : int
    this_sequence_header_bytes : :py:class:`bytearray`
        The bitstream byte-offset and raw bytes for the offending
        sequence_header in the stream.
    """

    def __init__(
        self,
        last_sequence_header_offset,
        last_sequence_header_bytes,
        this_sequence_header_offset,
        this_sequence_header_bytes,
    ):
        self.last_sequence_header_offset = last_sequence_header_offset
        self.last_sequence_header_bytes = last_sequence_header_bytes
        self.this_sequence_header_offset = this_sequence_header_offset
        self.this_sequence_header_bytes = this_sequence_header_bytes
        super(SequenceHeaderChangedMidSequence, self).__init__()

    def explain(self):
        first_difference_offset = 0
        for old, new in zip(
            bytearray(self.last_sequence_header_bytes),
            bytearray(self.this_sequence_header_bytes),
        ):
            if old != new:
                for i in reversed(range(8)):
                    if (old & (1 << i)) != (new & (1 << i)):
                        first_difference_offset += 7 - i
                        break
                break
            else:
                first_difference_offset += 8

        return """
            Sequence header is not byte-for-byte identical to the previous
            sequence header in the same sequence (11.1).

            The previous sequence header begins at bit offset {} and the
            current sequence header begins at bit offset {}.

            This sequence header differs from its predecessor starting with bit
            {}. That is, bit offset {} in the previous sequence header is
            different to bit offset {} in the current sequence header.

            Did the video format change without beginning a new sequence?

            Did the sequence header attempt to encode the same parameters in a
            different way (e.g. switching to a custom value rather than an
            equivalent preset)?
        """.format(
            to_bit_offset(self.last_sequence_header_offset),
            to_bit_offset(self.this_sequence_header_offset),
            first_difference_offset,
            to_bit_offset(self.last_sequence_header_offset) + first_difference_offset,
            to_bit_offset(self.this_sequence_header_offset) + first_difference_offset,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the previous sequence header

                {{cmd}} {{file}} --from-offset {} --to-offset {}

            To view the current sequence header

                {{cmd}} {{file}} --from-offset {} --to-offset {}
        """.format(
            to_bit_offset(self.last_sequence_header_offset),
            to_bit_offset(
                self.last_sequence_header_offset + len(self.last_sequence_header_bytes)
            )
            - 1,
            to_bit_offset(self.this_sequence_header_offset),
            to_bit_offset(
                self.this_sequence_header_offset + len(self.this_sequence_header_bytes)
            )
            - 1,
        )

    def offending_offset(self):
        return to_bit_offset(self.this_sequence_header_offset)


class BadProfile(ConformanceError):
    """
    parse_parameters (11.2.3) has been given an unrecognised profile number.

    The exception argument will contain the received profile number.
    """

    def __init__(self, profile):
        self.profile = profile
        super(BadProfile, self).__init__()

    def explain(self):
        return """
            An invalid profile number, {}, was provided in the parse parameters
            (11.2.3).

            See (C.2) for the list of allowed profile numbers.

            Perhaps this bitstream conforms to an earlier or later version of
            the VC-2 standard?
        """.format(
            self.profile
        )


class BadLevel(ConformanceError):
    """
    parse_parameters (11.2.3) has been given an unrecognised level number.

    The exception argument will contain the received level number.
    """

    def __init__(self, level):
        self.level = level
        super(BadLevel, self).__init__()

    def explain(self):
        return """
            An invalid level number, {}, was provided in the parse parameters
            (11.2.3).

            See (C.3) or SMPTE ST 2042-2 'VC-2 Level Definitions' for details
            of the supported levels and their codes.

            Perhaps this bitstream conforms to an earlier or later version of
            the VC-2 standard?
        """.format(
            self.level
        )


class GenericInvalidSequence(ConformanceError):
    """
    The sequence of data units in the VC-2 sequence does not match the generic
    sequence structure specified in (10.4.1)

    The offending parse code will be given in :py:attr:`parse_code` if it does
    not match the expected sequence structure. If the sequence ends prematurely
    ``None`` will be passed instead.

    :py:attr:`expected_parse_codes` enumerates the parse codes which would have
    been allowed.  This may be ``None`` if any parse code would be permitted.

    :py:attr:`expected_end` is True if it would have been acceptable for the
    sequence to have ended at this point.
    """

    def __init__(self, parse_code, expected_parse_codes, expected_end):
        self.parse_code = parse_code
        self.expected_parse_codes = expected_parse_codes
        self.expected_end = expected_end
        super(GenericInvalidSequence, self).__init__()

    def explain(self):
        return """
            The current sequence does not match the structure defined for VC-2
            sequences in (10.4.1).

            {}

            Did the sequence begin with a non-sequence header data unit?
        """.format(
            explain_parse_code_sequence_structure_restrictions(
                self.parse_code,
                self.expected_parse_codes,
                self.expected_end,
            )
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {cmd} {file} --from-offset 0 --to-offset {offset} --show parse_info
        """


class LevelInvalidSequence(ConformanceError):
    """
    The sequence of data units in the VC-2 sequence does not match the
    sequence structure specified for the current level (C.3).

    The offending parse code will be given in :py:attr:`parse_code` if it does
    not match the expected sequence structure. If the sequence ends prematurely
    ``None`` will be passed instead.

    :py:attr:`expected_parse_codes` enumerates the parse codes which would have
    been allowed.  This may be ``None`` if any parse code would be permitted.

    :py:attr:`expected_end` is True if it would have been acceptable for the
    sequence to have ended at this point.

    :py:attr:`level` is the current VC-2 level.
    """

    def __init__(self, parse_code, expected_parse_codes, expected_end, level):
        self.parse_code = parse_code
        self.expected_parse_codes = expected_parse_codes
        self.expected_end = expected_end
        self.level = level
        super(LevelInvalidSequence, self).__init__()

    def explain(self):
        return """
            The current sequence does not match the structure required by the
            current level, {}, ({}).

            {}

            {}
        """.format(
            self.level,
            LEVELS[self.level].standard,
            LEVEL_SEQUENCE_RESTRICTIONS[self.level].sequence_restriction_explanation,
            explain_parse_code_sequence_structure_restrictions(
                self.parse_code,
                self.expected_parse_codes,
                self.expected_end,
            ),
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {cmd} {file} --from-offset 0 --to-offset {offset} --show parse_info
        """


class ParseCodeNotAllowedInProfile(ConformanceError):
    """
    The parse code encountered is not allowed in the current profile (C.2).

    The offending parse_code and profile combination will be provided in
    :py:attr:`parse_code` and :py;attr:`profile`.
    """

    def __init__(self, parse_code, profile):
        self.parse_code = parse_code
        self.profile = profile
        super(ParseCodeNotAllowedInProfile, self).__init__()

    def explain(self):
        return """
            The parse code {} is not allowed in the {} profile (C.2).
        """.format(
            known_parse_code_to_string(self.parse_code),
            known_profile_to_string(self.profile),
        )


class ValueNotAllowedInLevel(ConformanceError):
    """
    A value was encountered in the bitstream which was not allowed by the
    current level. See
    :py:data:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS` for details
    of the constrained keys.

    :py:attr:`level_constrained_values` will be the previously specified
    values which were allowed by the level constraints.

    The offending key and value will be placed in :py:attr:`key` and
    :py:attr:`value` respectively.

    The allowed :py:class:`~vc2_conformance.constraint_table.ValueSet` will be
    placed in :py:attr:`allowed_values`.
    """

    def __init__(self, level_constrained_values, key, value, allowed_values):
        self.level_constrained_values = level_constrained_values
        self.key = key
        self.value = value
        self.allowed_values = allowed_values
        super(ValueNotAllowedInLevel, self).__init__()

    def explain(self):
        level = self.level_constrained_values["level"]

        return dedent(
            """
            The {} value {} is not allowed by level {}, expected {} ({}).

            The restriction above may be more constrained than expected due to
            one of the following previously encountered options:

            {}
        """
        ).format(
            self.key,
            self.value,
            level,
            str(self.allowed_values),
            LEVELS[level].standard,
            "\n".join(
                "* {} = {!r}".format(key, value)
                for key, value in self.level_constrained_values.items()
            ),
        )


class BadBaseVideoFormat(ConformanceError):
    """
    sequence_header (11.1) has been given an unrecognised base video format
    number.

    The exception argument will contain the received base video format number.
    """

    def __init__(self, base_video_format):
        self.base_video_format = base_video_format
        super(BadBaseVideoFormat, self).__init__()

    def explain(self):
        return """
            An invalid base video format, {}, was provided in a sequence header
            (11.3).

            See (Table 11.1) for a list of allowed video format numbers.

            Perhaps this bitstream conforms to an earlier or later version of
            the VC-2 standard?
        """.format(
            self.base_video_format
        )


class BadPictureCodingMode(ConformanceError):
    """
    sequence_header (11.1) has been given an unrecognised picture coding mode.

    The exception argument will contain the received picture coding mode value
    """

    def __init__(self, picture_coding_mode):
        self.picture_coding_mode = picture_coding_mode
        super(BadPictureCodingMode, self).__init__()

    def explain(self):
        return """
            An invalid picture coding mode, {:d}, was provided in a sequence
            header (11.5).

            See (11.5) for an enumeration of allowed values.
        """.format(
            self.picture_coding_mode
        )


class ZeroPixelFrameSize(ConformanceError):
    """
    (11.4.3) A custom frame size with a zero width or height was specified.

    The actual dimensions are specified as arguments.
    """

    def __init__(self, frame_width, frame_height):
        self.frame_width = frame_width
        self.frame_height = frame_height
        super(ZeroPixelFrameSize, self).__init__()

    def explain(self):
        return """
            An invalid custom frame size, {}x{} was provided containing zero
            pixels (11.4.3).
        """.format(
            self.frame_width, self.frame_height
        )


class BadColorDifferenceSamplingFormat(ConformanceError):
    """
    color_diff_sampling_format (11.4.4) has been given an unrecognised color
    difference format index.

    The exception argument will contain the index.
    """

    def __init__(self, color_diff_format_index):
        self.color_diff_format_index = color_diff_format_index
        super(BadColorDifferenceSamplingFormat, self).__init__()

    def explain(self):
        return """
            An invalid color difference sampling format, {:d}, was provided
            (11.4.4).

            See (Table 11.2) for an enumeration of allowed values.
        """.format(
            self.color_diff_format_index
        )


class BadSourceSamplingMode(ConformanceError):
    """
    scan_format (11.4.5) has been given an unrecognised source sampling mode.

    The exception argument will contain the offending mode.
    """

    def __init__(self, source_sampling):
        self.source_sampling = source_sampling
        super(BadSourceSamplingMode, self).__init__()

    def explain(self):
        return """
            An invalid source sampling mode, {:d}, was provided (11.4.5).

            See (11.4.5) for an enumeration of allowed values.
        """.format(
            self.source_sampling
        )


class BadPresetFrameRateIndex(ConformanceError):
    """
    frame_rate (11.4.6) has been given an unrecognised preset frame rate index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetFrameRateIndex, self).__init__()

    def explain(self):
        return """
            An invalid preset frame rate index, {:d}, was provided (11.4.6).

            See (Table 11.3) for an enumeration of allowed values.
        """.format(
            self.index
        )


class FrameRateHasZeroNumerator(ConformanceError):
    """
    (11.4.6) specifies that custom frame rates must not be zero (i.e. have a
    zero numerator)

    The denominator is provided as an argument.
    """

    def __init__(self, frame_rate_denom):
        self.frame_rate_denom = frame_rate_denom
        super(FrameRateHasZeroNumerator, self).__init__()

    def explain(self):
        return """
            An invalid frame rate, 0/{} fps, was provided (11.4.6).

            Frame rates must not be zero.
        """.format(
            self.frame_rate_denom
        )


class FrameRateHasZeroDenominator(ConformanceError):
    """
    (11.4.6) specifies that custom frame rates must not have zero in the
    denominator.

    The numerator is provided as an argument.
    """

    def __init__(self, frame_rate_numer):
        self.frame_rate_numer = frame_rate_numer
        super(FrameRateHasZeroDenominator, self).__init__()

    def explain(self):
        return """
            An invalid frame rate, {}/0 fps, was provided (11.4.6).

            The frame rate specification contains a division by zero.
        """.format(
            self.frame_rate_numer
        )


class BadPresetPixelAspectRatio(ConformanceError):
    """
    pixel_aspect_ratio_index (11.4.7) has been given an unrecognised preset
    index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetPixelAspectRatio, self).__init__()

    def explain(self):
        return """
            An invalid preset pixel aspect ratio index, {:d}, was provided
            (11.4.7).

            See (Table 11.4) for an enumeration of allowed values.
        """.format(
            self.index
        )


class PixelAspectRatioContainsZeros(ConformanceError):
    """
    (11.4.7) specifies that custom pixel aspect ratios must not have zeros.

    The ratio numerator/denominator is provided.
    """

    def __init__(self, pixel_aspect_ratio_numer, pixel_aspect_ratio_denom):
        self.pixel_aspect_ratio_numer = pixel_aspect_ratio_numer
        self.pixel_aspect_ratio_denom = pixel_aspect_ratio_denom
        super(PixelAspectRatioContainsZeros, self).__init__()

    def explain(self):
        return """
            An invalid pixel aspect ratio, {}:{}, was provided (11.4.7).

            Pixel aspect ratios must be valid ratios (i.e. not contain zeros).
        """.format(
            self.pixel_aspect_ratio_numer,
            self.pixel_aspect_ratio_denom,
        )


class CleanAreaOutOfRange(ConformanceError):
    """
    clean_area (11.4.8) specifies a clean area which goes beyond the boundaries
    of the picture.

    The offending clean area width, height, left and top offset will be
    included as exception arguments, followed by the picture width and height.
    """

    def __init__(
        self,
        clean_width,
        clean_height,
        left_offset,
        top_offset,
        frame_width,
        frame_height,
    ):
        self.clean_width = clean_width
        self.clean_height = clean_height
        self.left_offset = left_offset
        self.top_offset = top_offset
        self.frame_width = frame_width
        self.frame_height = frame_height
        super(CleanAreaOutOfRange, self).__init__()

    def explain(self):
        offending_dimensions = []
        if self.clean_width + self.left_offset > self.frame_width:
            offending_dimensions.append("width")
        if self.clean_height + self.top_offset > self.frame_height:
            offending_dimensions.append("height")

        return """
            The clean area {} extend{} beyond the frame boundary (11.4.8).

            * video_parameters[frame_width] = {}
            * left_offset ({}) + clean_width ({}) = {}
            * video_parameters[frame_height] = {}
            * top_offset ({}) + clean_height ({}) = {}

            Has a custom frame size been used and the clean area not been
            updated to match?
        """.format(
            " and ".join(offending_dimensions),
            ("s" if len(offending_dimensions) == 1 else ""),
            self.frame_width,
            self.left_offset,
            self.clean_width,
            self.left_offset + self.clean_width,
            self.frame_height,
            self.top_offset,
            self.clean_height,
            self.top_offset + self.clean_height,
        )


class BadCustomSignalExcursion(ConformanceError):
    """
    signal_range (11.4.9) requires that signal excursions must be at least 1.

    A string "luma" or "color_diff" followed by the offending excursion value
    should be provided as arguments.
    """

    def __init__(
        self,
        component_type_name,
        excursion,
    ):
        self.component_type_name = component_type_name
        self.excursion = excursion
        super(BadCustomSignalExcursion, self).__init__()

    def explain(self):
        return """
            Custom signal range {}_excursion must be at least 1 but {} was
            provided (11.4.9).
        """.format(
            self.component_type_name,
            self.excursion,
        )


class BadPresetSignalRange(ConformanceError):
    """
    signal_range (11.4.9) has been given an unrecognised preset index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetSignalRange, self).__init__()

    def explain(self):
        return """
            An invalid preset signal range index, {:d}, was provided (11.4.9).

            See (Table 11.5) for an enumeration of allowed values.
        """.format(
            self.index
        )


class BadPresetColorSpec(ConformanceError):
    """
    color_spec (11.4.10.1) has been given an unrecognised preset index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetColorSpec, self).__init__()

    def explain(self):
        return """
            An invalid preset color spec index, {:d}, was provided (11.4.10.1).

            See (Table 11.6) for an enumeration of allowed values.
        """.format(
            self.index
        )


class BadPresetColorPrimaries(ConformanceError):
    """
    color_primaries (11.4.10.2) has been given an unrecognised preset index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetColorPrimaries, self).__init__()

    def explain(self):
        return """
            An invalid color primaries index, {:d}, was provided (11.4.10.2).

            See (Table 11.7) for an enumeration of allowed values.
        """.format(
            self.index
        )


class BadPresetColorMatrix(ConformanceError):
    """
    color_matrix (11.4.10.3) has been given an unrecognised preset index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetColorMatrix, self).__init__()

    def explain(self):
        return """
            An invalid color matrix index, {:d}, was provided (11.4.10.3).

            See (Table 11.8) for an enumeration of allowed values.
        """.format(
            self.index
        )


class BadPresetTransferFunction(ConformanceError):
    """
    transfer_function (11.4.10.4) has been given an unrecognised preset index

    The exception argument will contain the offending index.
    """

    def __init__(self, index):
        self.index = index
        super(BadPresetTransferFunction, self).__init__()

    def explain(self):
        return """
            An invalid transfer function index, {:d}, was provided (11.4.10.4).

            See (Table 11.9) for an enumeration of allowed values.
        """.format(
            self.index
        )


class PictureDimensionsNotMultipleOfFrameDimensions(ConformanceError):
    """
    (11.6.2) specifies that the picture dimensions (luma_width, luma_height,
    color_diff_width and color_diff_height) must be a whole factor of the
    frame dimensions (i.e. frame_width and frame_height).

    The actual dimensions are specified as arguments.
    """

    def __init__(
        self,
        luma_width,
        luma_height,
        color_diff_width,
        color_diff_height,
        frame_width,
        frame_height,
    ):
        self.luma_width = luma_width
        self.luma_height = luma_height
        self.color_diff_width = color_diff_width
        self.color_diff_height = color_diff_height
        self.frame_width = frame_width
        self.frame_height = frame_height
        super(PictureDimensionsNotMultipleOfFrameDimensions, self).__init__()

    def explain(self):
        (
            luma_width_message,
            luma_height_message,
            color_diff_width_message,
            color_diff_height_message,
        ) = (
            " (not a factor of {})".format(frame_dimen)
            if component_dimen == 0 or (frame_dimen % component_dimen) != 0
            else ""
            for (component_dimen, frame_dimen) in [
                (self.luma_width, self.frame_width),
                (self.luma_height, self.frame_height),
                (self.color_diff_width, self.frame_width),
                (self.color_diff_height, self.frame_height),
            ]
        )

        return """
            The frame dimensions cannot be evenly divided by the current color
            difference sampling format and picture coding mode (11.6.2)

            Frame dimensions:

            * frame_width: {}
            * frame_height: {}

            The dimensions computed by picture_dimensions were:

            * luma_width: {}{}
            * luma_height: {}{}
            * color_diff_width: {}{}
            * color_diff_height: {}{}

            Was a frame size with an odd width or height used along with a
            non-4:4:4 color difference sampling mode or when pictures are
            fields?

            Was the source sampling mode (11.4.5) used instead of the picture
            coding mode (11.5) to determine the picture size?
        """.format(
            self.frame_width,
            self.frame_height,
            self.luma_width,
            luma_width_message,
            self.luma_height,
            luma_height_message,
            self.color_diff_width,
            color_diff_width_message,
            self.color_diff_height,
            color_diff_height_message,
        )


class NonConsecutivePictureNumbers(ConformanceError):
    """
    (12.2) and (14.2) Picture numbers for each picture must contain consecutive
    picture numbers (wrapping at 2**32).

    :py:attr:`last_picture_number_offset` contains the (byte_offset, next_bit)
    offset of the previous picture number in the sequence.

    :py:attr:`last_picture_number` contains the picture number of the previous
    picture in the sequence.

    :py:attr:`picture_number_offset` contains the (byte_offset, next_bit)
    offset of the offending picture number in the sequence.

    :py:attr:`picture_number` contains the picture number of the offending
    picture in the sequence.
    """

    def __init__(
        self,
        last_picture_number_offset,
        last_picture_number,
        picture_number_offset,
        picture_number,
    ):
        self.last_picture_number_offset = last_picture_number_offset
        self.last_picture_number = last_picture_number
        self.picture_number_offset = picture_number_offset
        self.picture_number = picture_number
        super(NonConsecutivePictureNumbers, self).__init__()

    def explain(self):
        return """
            Non-consecutive picture number, got {} after {} (12.2) and (14.2).

            Picture numbers must have consecutive, ascending integer values,
            wrapping at (2**32)-1 back to 0.

            * Previous picture number defined at bit offset {}
            * Current picture number defined at bit offset {}

            Was this picture taken from another sequence without being assigned
            a new picture number?
        """.format(
            self.picture_number,
            self.last_picture_number,
            to_bit_offset(*self.last_picture_number_offset),
            to_bit_offset(*self.picture_number_offset),
        )

    def bitstream_viewer_hint(self):
        return """
            To view the erroneous picture number definition:

                {{cmd}} {{file}} --offset {}

            To view the previous picture number definition:

                {{cmd}} {{file}} --offset {}
        """.format(
            to_bit_offset(*self.picture_number_offset),
            to_bit_offset(*self.last_picture_number_offset),
        )


class OddNumberOfFieldsInSequence(ConformanceError):
    """
    (10.4.3) When pictures are fields, a sequence must have a whole number of
    frames (i.e. an even number of fields).

    The actual number of fields/pictures in the offending sequence will be
    included as an argument.
    """

    def __init__(self, num_fields_in_sequence):
        self.num_fields_in_sequence = num_fields_in_sequence
        super(OddNumberOfFieldsInSequence, self).__init__()

    def explain(self):
        return """
            Sequence contains a non-whole number of frames ({} fields)
            (10.4.3).

            When pictures are fields, an even number of fields/pictures must be
            included in each sequence.

            Was the sequence truncated mid-frame?
        """.format(
            self.num_fields_in_sequence,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {cmd} {file} --from-offset 0 --to-offset {offset} --show picture_parse --show fragment_parse
        """


class EarliestFieldHasOddPictureNumber(ConformanceError):
    """
    (12.2) The earliest field of each frame must have an even picture number.

    The offending picture number will be included as an argument.
    """

    def __init__(self, picture_number):
        self.picture_number = picture_number
        super(EarliestFieldHasOddPictureNumber, self).__init__()

    def explain(self):
        return """
            First field in sequence has an odd picture number, {} (12.2).

            When pictures are fields, the earliest field/picture in each frame
            in the sequence must have an even picture number.

            Was the sequence truncated mid-frame?
        """.format(
            self.picture_number,
        )


class BadWaveletIndex(ConformanceError):
    """
    transform_parameters (12.4.1) has been given an unrecognised wavelet index.

    The exception argument will contain the offending wavelet index
    """

    def __init__(self, wavelet_index):
        self.wavelet_index = wavelet_index
        super(BadWaveletIndex, self).__init__()

    def explain(self):
        return """
            An invalid wavelet index, {:d}, was provided in the transform
            parameters (12.4.1).

            See (Table 12.1) for an enumeration of allowed values.
        """.format(
            self.wavelet_index
        )


class BadHOWaveletIndex(ConformanceError):
    """
    extended_transform_parameters (12.4.4.2) has been given an unrecognised
    wavelet index.

    The exception argument will contain the offending wavelet index
    """

    def __init__(self, wavelet_index_ho):
        self.wavelet_index_ho = wavelet_index_ho
        super(BadHOWaveletIndex, self).__init__()

    def explain(self):
        return """
            An invalid horizontal only wavelet index, {:d}, was provided in the
            extended transform parameters (12.4.4.2).

            See (Table 12.1) for an enumeration of allowed values.
        """.format(
            self.wavelet_index_ho
        )


class ZeroSlicesInCodedPicture(ConformanceError):
    """
    (12.4.5.2) slice_parameters must not allow either slice count to be zero.

    The exception argument will contain the offending slice counts
    """

    def __init__(self, slices_x, slices_y):
        self.slices_x = slices_x
        self.slices_y = slices_y
        super(ZeroSlicesInCodedPicture, self).__init__()

    def explain(self):
        return """
            Invalid slice count, {}x{}, specified in slice parameters
            (12.4.5.2).

            There must be at least one slice in either dimension.
        """.format(
            self.slices_x,
            self.slices_y,
        )


class SliceBytesHasZeroDenominator(ConformanceError):
    """
    (12.4.5.2) specifies that slice_bytes_denominator must not be zero (to
    avoid division by zero)

    The numerator is provided as an argument.
    """

    def __init__(self, slice_bytes_numerator):
        self.slice_bytes_numerator = slice_bytes_numerator
        super(SliceBytesHasZeroDenominator, self).__init__()

    def explain(self):
        return """
            Invalid slice bytes count, {}/0, in slice parameters (12.4.5.2).

            Division by zero.
        """.format(
            self.slice_bytes_numerator,
        )


class SliceBytesIsLessThanOne(ConformanceError):
    """
    (12.4.5.2) specifies that slice_bytes_numerator/slice_bytes_denominator
    must be greater or equal to one byte.

    The offending numerator and denominator is provided as an argument.
    """

    def __init__(self, slice_bytes_numerator, slice_bytes_denominator):
        self.slice_bytes_numerator = slice_bytes_numerator
        self.slice_bytes_denominator = slice_bytes_denominator
        super(SliceBytesIsLessThanOne, self).__init__()

    def explain(self):
        return """
            Slice bytes count, {}/{}, in slice parameters is less than one
            (12.4.5.2).

            Slices must be at least 1 byte.
        """.format(
            self.slice_bytes_numerator,
            self.slice_bytes_denominator,
        )


class NoQuantisationMatrixAvailable(ConformanceError):
    """
    (12.4.5.3) specifies that custom quantisation matrices must be used in
    cases where a default is not defined by the standard.

    The offending combination of wavelet_index, wavelet_index_ho, dwt_depth,
    and dwt_depth_ho are provided as arguments.
    """

    def __init__(self, wavelet_index, wavelet_index_ho, dwt_depth, dwt_depth_ho):
        self.wavelet_index = wavelet_index
        self.wavelet_index_ho = wavelet_index_ho
        self.dwt_depth = dwt_depth
        self.dwt_depth_ho = dwt_depth_ho
        super(NoQuantisationMatrixAvailable, self).__init__()

    def explain(self):
        return """
            A default quantisation matrix is not available for current
            transform and no custom quantisation matrix has been supplied
            (12.4.5.3).

            The current transform is defined as:

            * wavelet_index = {}
            * dwt_depth = {}
            * wavelet_index_ho = {}
            * dwt_depth_ho = {}
        """.format(
            known_wavelet_to_string(self.wavelet_index),
            self.dwt_depth,
            known_wavelet_to_string(self.wavelet_index_ho),
            self.dwt_depth_ho,
        )


class QuantisationMatrixValueNotAllowedInLevel(ConformanceError):
    """
    (C.3) A custom quantisation matrix value was used which was not allowed by the
    current level.

    The value and expected values (as a
    :py:class:`~vc2_conformance.constraint_table.ValueSet`) will be passed as
    arguments along with the current level constraint dictionary.
    """

    def __init__(self, value, allowed_values, level_constrained_values):
        self.value = value
        self.allowed_values = allowed_values
        self.level_constrained_values = level_constrained_values
        super(QuantisationMatrixValueNotAllowedInLevel, self).__init__()

    def explain(self):
        level = self.level_constrained_values["level"]

        return dedent(
            """
            Custom quantisation matrix contains a value, {}, outside the
            range {} allowed by the current level, {} ({}).

            The restriction above may be more constrained than expected due to
            one of the following previously encountered options:

            {}
        """
        ).format(
            self.value,
            self.allowed_values,
            level,
            LEVELS[level].standard,
            "\n".join(
                "* {} = {!r}".format(key, value)
                for key, value in self.level_constrained_values.items()
            ),
        )


class SliceSizeScalerIsZero(ConformanceError):
    """
    (12.4.5.2) A slice_size_scaler value of zero was given.
    """

    def explain(self):
        return """
            Slice parameter slice_size_scaler must not be zero (12.4.5.2)
        """


class InvalidSliceYLength(ConformanceError):
    """
    (13.5.3.1) ld_slice must have its slice_y_length value be within the length
    of the whole slice.

    The offending slice_y_length, the number of bytes allowed in the slice and
    the slice coordinates are provided as an argument.
    """

    def __init__(self, slice_y_length, slice_bytes, sx, sy):
        self.slice_y_length = slice_y_length
        self.slice_bytes = slice_bytes
        self.sx = sx
        self.sy = sy
        super(InvalidSliceYLength, self).__init__()

    def explain(self):
        slice_bits_left = 8 * self.slice_bytes

        # Account for qindex
        slice_bits_left -= 7

        # Account for length field
        length_bits = intlog2(slice_bits_left)
        slice_bits_left -= length_bits

        return """
            Low-delay slice_y_length value, {slice_y_length}, is out of range,
            expected a value no greater than {max_slice_y_length} (13.5.3.1).

            * The current slice (sx={sx}, sy={sy}) is {slice_bytes} bytes
              ({slice_bits} bits) long (see slice_bytes() (13.5.3.2)).
            * 7 bits are reserved for the qindex field.
            * intlog2({slice_minus_qindex_bits}) = {length_bits} bits are
              reserved for the slice_y_length field.
            * This leaves {max_slice_y_length} bits to split between the
              luminance and color difference components.

            Was the size of this slice correctly computed?

            Were the size of the qindex and slice_y_length fields accounted
            for?
        """.format(
            slice_y_length=self.slice_y_length,
            max_slice_y_length=slice_bits_left,
            sx=self.sx,
            sy=self.sy,
            slice_bytes=self.slice_bytes,
            slice_bits=self.slice_bytes * 8,
            slice_minus_qindex_bits=(self.slice_bytes * 8) - 7,
            length_bits=length_bits,
        )


class FragmentedPictureRestarted(ConformanceError):
    """
    (14.2) Not all of the slices in a fragmented picture arrived before a new
    fragment with fragment_slice_count==0 arrived.

    The (byte_offset, next_bit_offset) offset of the first fragment header in
    the incomplete picture and of this fragment header are included as
    arguments along with the number of slices received and remaining.
    """

    def __init__(
        self,
        initial_fragment_offset,
        this_fragment_offset,
        fragment_slices_received,
        fragment_slices_remaining,
    ):
        self.initial_fragment_offset = initial_fragment_offset
        self.this_fragment_offset = this_fragment_offset
        self.fragment_slices_received = fragment_slices_received
        self.fragment_slices_remaining = fragment_slices_remaining
        super(FragmentedPictureRestarted, self).__init__()

    def explain(self):
        return """
            A picture fragment with fragment_slice_count=0 was encountered
            while {} slice{} still outstanding (14.2).

            The previous fragmented picture started at bit offset {} and {} of
            {} expected slices were received before the current picture
            fragment with fragment_slice_count=0 arrived at bit offset {}.

            Was a picture fragment with fragment_slice_count=0 incorrectly used
            as padding while waiting for some picture slices to be ready?

            Were some picture fragments omitted when copying a fragmented
            picture from another sequence?
        """.format(
            self.fragment_slices_remaining,
            " is" if self.fragment_slices_remaining == 1 else "s are",
            to_bit_offset(*self.initial_fragment_offset),
            self.fragment_slices_received,
            self.fragment_slices_received + self.fragment_slices_remaining,
            to_bit_offset(*self.this_fragment_offset),
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {{cmd}} {{file}} --from-offset {} --to_offset {{offset}} --show fragment_parse --hide slice
        """.format(
            to_bit_offset(*self.initial_fragment_offset)
        )


class SequenceContainsIncompleteFragmentedPicture(ConformanceError):
    """
    (14.2) Sequences must not terminate mid-fragmented picture.

    The (byte_offset, next_bit_offset) offset of the first fragment header in
    the incomplete picture is included an argument along with the number of
    slices received and remaining.
    """

    def __init__(
        self,
        initial_fragment_offset,
        fragment_slices_received,
        fragment_slices_remaining,
    ):
        self.initial_fragment_offset = initial_fragment_offset
        self.fragment_slices_received = fragment_slices_received
        self.fragment_slices_remaining = fragment_slices_remaining
        super(SequenceContainsIncompleteFragmentedPicture, self).__init__()

    def explain(self):
        return """
            A sequence terminated while {} slice{} still outstanding in a
            fragmented picture (14.2).

            The fragmented picture started at bit offset {} and {} of
            {} expected slices were received before the end of the sequence was
            encountered.

            Were some picture fragments omitted when copying a fragmented
            picture from another sequence?
        """.format(
            self.fragment_slices_remaining,
            " is" if self.fragment_slices_remaining == 1 else "s are",
            to_bit_offset(*self.initial_fragment_offset),
            self.fragment_slices_received,
            self.fragment_slices_received + self.fragment_slices_remaining,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {{cmd}} {{file}} --from-offset {} --to_offset {{offset}} --show parse_info --show fragment_parse --hide slice
        """.format(
            to_bit_offset(*self.initial_fragment_offset)
        )


class PictureInterleavedWithFragmentedPicture(ConformanceError):
    """
    (14.2) Picture data units may not be interleaved with in-progress
    fragmented pictures.

    The (byte_offset, next_bit_offset) offset of the first fragment header in
    the incomplete picture and the offending picture data unit's header is
    included an argument along with the number of slices received and remaining
    for the fragmented picture.
    """

    def __init__(
        self,
        initial_fragment_offset,
        this_offset,
        fragment_slices_received,
        fragment_slices_remaining,
    ):
        self.initial_fragment_offset = initial_fragment_offset
        self.this_offset = this_offset
        self.fragment_slices_received = fragment_slices_received
        self.fragment_slices_remaining = fragment_slices_remaining
        super(PictureInterleavedWithFragmentedPicture, self).__init__()

    def explain(self):
        return """
            A non-fragmented picture was encountered while {} slice{} still
            outstanding in a fragmented picture (14.2).

            The fragmented picture started at bit offset {} and {} of {}
            expected slices were received before the non-fragmented picture was
            encountered at bit offset {}.

            Were some picture fragments omitted when copying a fragmented
            picture from another sequence?
        """.format(
            self.fragment_slices_remaining,
            " is" if self.fragment_slices_remaining == 1 else "s are",
            to_bit_offset(*self.initial_fragment_offset),
            self.fragment_slices_received,
            self.fragment_slices_received + self.fragment_slices_remaining,
            to_bit_offset(*self.this_offset),
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {{cmd}} {{file}} --from-offset {} --to_offset {{offset}} --show parse_info --show fragment_parse --show picture_parse --hide slice
        """.format(
            to_bit_offset(*self.initial_fragment_offset)
        )


class PictureNumberChangedMidFragmentedPicture(ConformanceError):
    """
    (14.2) Picture numbers in fragment_headers which are part of the same
    fragmented picture must be identical.

    :py:attr:`last_picture_number_offset` contains the (byte_offset, next_bit)
    offset of the previous picture number in the sequence.

    :py:attr:`last_picture_number` contains the picture number of the previous
    fragment in the sequence.

    :py:attr:`picture_number_offset` contains the (byte_offset, next_bit)
    offset of the offending picture number in the sequence.

    :py:attr:`picture_number` contains the picture number of the offending
    fragment in the sequence.
    """

    def __init__(
        self,
        last_picture_number_offset,
        last_picture_number,
        picture_number_offset,
        picture_number,
    ):
        self.last_picture_number_offset = last_picture_number_offset
        self.last_picture_number = last_picture_number
        self.picture_number_offset = picture_number_offset
        self.picture_number = picture_number
        super(PictureNumberChangedMidFragmentedPicture, self).__init__()

    def explain(self):
        return """
            The picture number changed from {} to {} within the same fragmented
            picture (14.2).

            The previous fragment in this fragmented picture defined its
            picture number at bit offset {}. The current fragment provided a
            different picture number at bit offset {}.

            Was the picture number incremented for every fragment rather than
            for every complete picture in the stream?
        """.format(
            self.last_picture_number,
            self.picture_number,
            to_bit_offset(*self.last_picture_number_offset),
            to_bit_offset(*self.picture_number_offset),
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {{cmd}} {{file}} --from-offset {} --to_offset {{offset}} --show fragment_parse --hide slice
        """.format(
            to_bit_offset(*self.last_picture_number_offset),
        )


class TooManySlicesInFragmentedPicture(ConformanceError):
    """
    (14.2) A fragmented picture must not contain more slices than necessary.

    The (byte_offset, next_bit_offset) offset of the first fragment header in
    the picture and the offending fragment's header is
    included an argument along with the number of slices received, remaining
    and included in the offending fragment.
    """

    def __init__(
        self,
        initial_fragment_offset,
        this_fragment_offset,
        fragment_slices_received,
        fragment_slices_remaining,
        fragment_slice_count,
    ):
        self.initial_fragment_offset = initial_fragment_offset
        self.this_fragment_offset = this_fragment_offset
        self.fragment_slices_received = fragment_slices_received
        self.fragment_slices_remaining = fragment_slices_remaining
        self.fragment_slice_count = fragment_slice_count
        super(TooManySlicesInFragmentedPicture, self).__init__()

    def explain(self):
        total_slices = self.fragment_slices_received + self.fragment_slices_remaining

        return """
            The current fragmented picture contains too many picture slices
            (14.2).

            This fragmented picture (starting at bit offset {}) consists of a
            total of {} picture slices. {} slices have already been received
            but the current fragment (at bit offset {}) contains {} slices
            while only {} more are expected.
        """.format(
            to_bit_offset(*self.initial_fragment_offset),
            total_slices,
            self.fragment_slices_received,
            to_bit_offset(*self.this_fragment_offset),
            self.fragment_slice_count,
            self.fragment_slices_remaining,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {{cmd}} {{file}} --from-offset {} --to_offset {{offset}} --show fragment_parse --hide slice
        """.format(
            to_bit_offset(*self.initial_fragment_offset),
        )


class FragmentSlicesNotContiguous(ConformanceError):
    """
    (14.2) A fragmented picture must contain every slice in the picture exactly
    once, provided in raster-scan order and starting at (sx=0, sy=0).

    The (byte_offset, next_bit_offset) offset of the first fragment header in
    the picture and the offending fragment's header is included an argument
    along with the offending slice coordinates and expected slice coordinates.
    """

    def __init__(
        self,
        initial_fragment_offset,
        this_fragment_offset,
        fragment_x_offset,
        fragment_y_offset,
        expected_fragment_x_offset,
        expected_fragment_y_offset,
    ):
        self.initial_fragment_offset = initial_fragment_offset
        self.this_fragment_offset = this_fragment_offset
        self.fragment_x_offset = fragment_x_offset
        self.fragment_y_offset = fragment_y_offset
        self.expected_fragment_x_offset = expected_fragment_x_offset
        self.expected_fragment_y_offset = expected_fragment_y_offset
        super(FragmentSlicesNotContiguous, self).__init__()

    def explain(self):
        return """
            The current picture fragment's slices are non-contiguous (14.2).

            The fragmented picture starting at bit offset {} contains a
            fragment at bit offset {} with an unexpected start offset:

            * fragment_x_offset = {} (should be {})
            * fragment_y_offset = {} (should be {})

            Fragmented pictures must include picture slices in raster-scan
            order starting with sx=0, sy=0 and without leaving any gaps.
        """.format(
            to_bit_offset(*self.initial_fragment_offset),
            to_bit_offset(*self.this_fragment_offset),
            self.fragment_x_offset,
            self.expected_fragment_x_offset,
            self.fragment_y_offset,
            self.expected_fragment_y_offset,
        )

    def bitstream_viewer_hint(self):
        return """
            To view the offending part of the bitstream:

                {{cmd}} {{file}} --from-offset {} --to_offset {{offset}} --show fragment_parse --hide slice
        """.format(
            to_bit_offset(*self.initial_fragment_offset),
        )


class PresetFrameRateNotSupportedByVersion(ConformanceError):
    """
    preset_frame_rate (11.4.6) was given a value not supported by stream
    version specified specified (11.2.2).

    The exception argument will contain the offending index and specified
    version number.
    """

    def __init__(self, index, major_version):
        self.index = index
        self.major_version = major_version
        super(PresetFrameRateNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The preset frame rate index {:d} ({} FPS) (11.4.6) is only supported
            when major_version is at least {} but major_version is {}.  See
            (11.2.2).
        """.format(
            self.index,
            Fraction(*PRESET_FRAME_RATES[self.index]),
            preset_frame_rate_version_implication(self.index),
            self.major_version,
        )


class PresetSignalRangeNotSupportedByVersion(ConformanceError):
    """
    preset_signal_range (11.4.9) was given a value not supported by stream
    version specified specified (11.2.2).

    The exception argument will contain the offending index and specified
    version number.
    """

    def __init__(self, index, major_version):
        self.index = index
        self.major_version = major_version
        super(PresetSignalRangeNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The preset signal range index {:d} ({}) (11.4.9) is only supported
            when major_version is at least {} but major_version is {}. See
            (11.2.2).
        """.format(
            self.index,
            PresetSignalRanges(self.index).name,
            preset_signal_range_version_implication(self.index),
            self.major_version,
        )


class PresetColorSpecNotSupportedByVersion(ConformanceError):
    """
    preset_color_spec (11.4.10.1) was given a value not supported by stream
    version specified specified (11.2.2).

    The exception argument will contain the offending index and specified
    version number.
    """

    def __init__(self, index, major_version):
        self.index = index
        self.major_version = major_version
        super(PresetColorSpecNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The preset color spec index {:d} ({}) (11.4.10.1) is only supported
            when major_version is at least {} but major_version is {}.  See
            (11.2.2).
        """.format(
            self.index,
            PresetColorSpecs(self.index).name,
            preset_color_spec_version_implication(self.index),
            self.major_version,
        )


class PresetColorPrimariesNotSupportedByVersion(ConformanceError):
    """
    preset_color_primaries (11.4.10.2) was given a value not supported by
    stream version specified specified (11.2.2).

    The exception argument will contain the offending index and specified
    version number.
    """

    def __init__(self, index, major_version):
        self.index = index
        self.major_version = major_version
        super(PresetColorPrimariesNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The preset color primaries index {:d} ({}) (11.4.10.2) is only
            supported when major_version is at least {} but major_version is
            {}. See (11.2.2).
        """.format(
            self.index,
            PresetColorPrimaries(self.index).name,
            preset_color_primaries_version_implication(self.index),
            self.major_version,
        )


class PresetColorMatrixNotSupportedByVersion(ConformanceError):
    """
    preset_color_matrix (11.4.10.3) was given a value not supported by
    stream version specified specified (11.2.2).

    The exception argument will contain the offending index and specified
    version number.
    """

    def __init__(self, index, major_version):
        self.index = index
        self.major_version = major_version
        super(PresetColorMatrixNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The preset color matrix index {:d} ({}) (11.4.10.3) is only
            supported when major_version is at least {} but major_version is
            {}. See (11.2.2).
        """.format(
            self.index,
            PresetColorMatrices(self.index).name,
            preset_color_matrix_version_implication(self.index),
            self.major_version,
        )


class PresetTransferFunctionNotSupportedByVersion(ConformanceError):
    """
    preset_transfer_function (11.4.10.4) was given a value not supported by
    stream version specified specified (11.2.2).

    The exception argument will contain the offending index and specified
    version number.
    """

    def __init__(self, index, major_version):
        self.index = index
        self.major_version = major_version
        super(PresetTransferFunctionNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The preset transfer function index {:d} ({}) (11.4.10.4) is only
            supported when major_version is at least {} but major_version is
            {}. See (11.2.2).
        """.format(
            self.index,
            PresetTransferFunctions(self.index).name,
            preset_transfer_function_version_implication(self.index),
            self.major_version,
        )


class ParseCodeNotSupportedByVersion(ConformanceError):
    """
    parse_code (10.5.1) specified a data unit type which is not supported by
    stream version specified specified (11.2.2).

    The exception argument will contain the parse code and specified version number.
    """

    def __init__(self, parse_code, major_version):
        self.parse_code = parse_code
        self.major_version = major_version
        super(ParseCodeNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The parse code (10.5.1) specifies a {} data unit but this is only
            supported when major_version is at least {} but major_version is
            {}. See (11.2.2).
        """.format(
            known_parse_code_to_string(self.parse_code),
            parse_code_version_implication(self.parse_code),
            self.major_version,
        )


class ProfileNotSupportedByVersion(ConformanceError):
    """
    parse_parameters (11.2.1) specified a profile which is not supported by
    stream version specified specified (11.2.2).

    The exception argument will contain the profile and specified version number.
    """

    def __init__(self, profile, major_version):
        self.profile = profile
        self.major_version = major_version
        super(ProfileNotSupportedByVersion, self).__init__()

    def explain(self):
        return """
            The {} profile is only supported when major_version is at least {}
            but major_version is {}. See (11.2.2).
        """.format(
            known_profile_to_string(self.profile),
            profile_version_implication(self.profile),
            self.major_version,
        )


class MajorVersionTooLow(ConformanceError):
    """
    Thrown when a sequence specifies a major_version (11.2.1) which is less
    than 1 as required by (11.2.2).

    The exception argument will contain the major_version given in the
    sequence.
    """

    def __init__(self, major_version):
        self.major_version = major_version
        super(MajorVersionTooLow, self).__init__()

    def explain(self):
        return """
            The major_version (11.2.1) must be at least {} but {} was given
            instead, see (11.2.2).
        """.format(
            MINIMUM_MAJOR_VERSION,
            self.major_version,
        )


class MinorVersionNotZero(ConformanceError):
    """
    Thrown when a sequence specifies a minor_version (11.2.1) which is not 0 as
    required by (11.2.2).

    The exception argument will contain the minor_version given in the
    sequence.
    """

    def __init__(self, minor_version):
        self.minor_version = minor_version
        super(MinorVersionNotZero, self).__init__()

    def explain(self):
        return """
            The minor_version (11.2.1) must be 0 but {} was given instead, see
            (11.2.2).
        """.format(
            self.minor_version,
        )


class MajorVersionTooHigh(ConformanceError):
    """
    Thrown when a sequence specifies a major_version (11.2.1) which is higher
    than required for the features actually used in the sequence.

    The exception argument will contain the major_version given in the
    sequence and the expected major_version.
    """

    def __init__(self, major_version, expected_major_version):
        self.major_version = major_version
        self.expected_major_version = expected_major_version
        super(MajorVersionTooHigh, self).__init__()

    def explain(self):
        return """
            The major_version (11.2.1) specified, {}, is too high: only
            features requiring major_version {} were used in the sequence, see
            (11.2.2).
        """.format(
            self.major_version,
            self.expected_major_version,
        )



"""
Tests which verify the extended transform parameters (12.4.4) are correctly
read.
"""

from copy import deepcopy

from vc2_data_tables import ParseCodes

from vc2_conformance.bitstream import (
    Stream,
    autofill_major_version,
)

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.constraint_table import allowed_values_for

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.codec_features import codec_features_to_trivial_level_constraints

from vc2_conformance.picture_generators import static_sprite

from vc2_conformance.encoder import make_sequence


def update_extended_transform_parameters(sequence, *args, **kwargs):
    r"""
    Apply a :py:meth:`dict.update` to all
    :py:class:`~vc2_conformance.bitstream.ExtendedTransformParameters` in a
    :py:class:`~vc2_conformance.bitstream.Sequence`.

    Returns two values:
    * A new sequence (a copy is made and the old sequence left unmodified)
    * A bool which is True if the changes specified changed the extended
      transform parameter block.
    """
    sequence = deepcopy(sequence)
    changed = False
    for data_unit in sequence["data_units"]:
        if data_unit["parse_info"]["parse_code"] in (
            ParseCodes.low_delay_picture,
            ParseCodes.high_quality_picture,
        ):
            extended_transform_parameters = data_unit["picture_parse"][
                "wavelet_transform"
            ]["transform_parameters"]["extended_transform_parameters"]
        elif (
            data_unit["parse_info"]["parse_code"]
            in (
                ParseCodes.low_delay_picture_fragment,
                ParseCodes.high_quality_picture_fragment,
            )
            and "transform_parameters" in data_unit["fragment_parse"]
        ):
            extended_transform_parameters = data_unit["fragment_parse"][
                "transform_parameters"
            ]["extended_transform_parameters"]
        else:
            continue

        before = extended_transform_parameters.copy()
        extended_transform_parameters.update(*args, **kwargs)
        changed |= extended_transform_parameters != before

    return sequence, changed


@decoder_test_case_generator
def extended_transform_parameters(codec_features):
    """
    **Tests that extended transform parameter flags are handled correctly.**

    Ensures that extended transform parameters fields (12.4.4) are correctly
    handled by decoders for symmetric transform modes.

    ``extended_transform_parameters[asym_transform_index_flag]``
        Verifies that ``asym_transform_index_flag`` can be set to ``1``.

    ``extended_transform_parameters[asym_transform_flag]``
        Verifies that ``asym_transform_flag`` can be set to ``1``.

    These test cases are skipped for streams whose major version is less than 3
    (which do not support the extended transform parameters header).
    Additionally, these test cases are skipped for asymmetric transforms when
    the flag being tested must already be ``1``.
    """
    # Generate a base sequence in which we'll replace the extended transform
    # parameters later
    base_sequence = make_sequence(
        codec_features,
        static_sprite(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
    )

    # Skip this test if the generated sequence does not use major_version 3
    # since no extended transform parameters field will be present.
    autofill_major_version(Stream(sequences=[base_sequence]))
    major_versions = [
        data_unit["sequence_header"]["parse_parameters"]["major_version"]
        for data_unit in base_sequence["data_units"]
        if data_unit["parse_info"]["parse_code"] == ParseCodes.sequence_header
    ]
    if major_versions[0] != 3:
        return

    # Try enabling the asym_transform_index_flag and asym_transform_flag, if
    # not already enabled and if the current level permits it
    constrained_values = codec_features_to_trivial_level_constraints(codec_features)
    if True in allowed_values_for(
        LEVEL_CONSTRAINTS, "asym_transform_index_flag", constrained_values
    ):
        sequence, changed = update_extended_transform_parameters(
            base_sequence,
            asym_transform_index_flag=True,
            wavelet_index_ho=codec_features["wavelet_index_ho"],
        )
        if changed:
            yield TestCase(Stream(sequences=[sequence]), "asym_transform_index_flag")

    if True in allowed_values_for(
        LEVEL_CONSTRAINTS, "asym_transform_flag", constrained_values
    ):
        sequence, changed = update_extended_transform_parameters(
            base_sequence,
            asym_transform_flag=True,
            dwt_depth_ho=codec_features["dwt_depth_ho"],
        )
        if changed:
            yield TestCase(Stream(sequences=[sequence]), "asym_transform_flag")



"""
Common logic relating to loading :py:mod:`vc2_bit_widths` data and assembling
test pictures.
"""

import os

from vc2_data_tables import QUANTISATION_MATRICES

from vc2_conformance.dimensions_and_depths import compute_dimensions_and_depths

from vc2_bit_widths.helpers import (
    quantisation_index_bound,
    evaluate_filter_bounds,
    evaluate_test_pattern_outputs,
    generate_test_pictures,
)

from vc2_bit_widths.bundle import (
    bundle_get_static_filter_analysis,
    bundle_get_optimised_synthesis_test_patterns,
)

from vc2_conformance_data import STATIC_FILTER_ANALYSIS_BUNDLE_FILENAME


__all__ = [
    "get_bundle_filename",
    "get_test_pictures",
]


class MissingStaticAnalysisError(KeyError):
    """
    Exception thrown when a static analysis is not available for the codec
    configuration used.
    """


def get_bundle_filename():
    """
    Get the filename to use for loading the VC-2 bit widths analysis bundle
    file.

    If the ``VC2_BIT_WIDTHS_BUNDLE`` environment variable is set, uses this as
    the bundle file. Otherwise, falls back to
    :py:data:`vc2_conformance_data.STATIC_FILTER_ANALYSIS_BUNDLE_FILENAME`
    which contains analyses for all codec configurations with a default
    quantisation matrix.
    """
    default_filename = STATIC_FILTER_ANALYSIS_BUNDLE_FILENAME

    # NB, Also uses the default if the environment variable is present but
    # empty
    return os.environ.get("VC2_BIT_WIDTHS_BUNDLE", "") or default_filename


def get_test_pictures(codec_features, bundle_filename=get_bundle_filename()):
    """
    Gets a set of test pictures for a codec with the specified codec
    configuration.

    Raises a :py:exc:`MissingStaticAnalysisError` if no static filter analysis
    is available for the specified codec.

    Parameters
    ==========
    bundle_filename : str
        Filename of the VC-2 bit widths bundle file to read test patterns from.
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`

    Returns
    =======
    analysis_luma_pictures : [:py:class:`AnalysisPicture`, ...]
    synthesis_luma_pictures : [:py:class:`SynthesisPicture`, ...]
    analysis_color_diff_pictures : [:py:class:`AnalysisPicture`, ...]
    synthesis_color_diff_pictures : [:py:class:`SynthesisPicture`, ...]
        Test pictures for luma and color difference components respectively.
        (See :py:func:`vc2_bit_widths.helpers.generate_test_pictures`.)
    """
    if codec_features["quantization_matrix"] is not None:
        quantisation_matrix = codec_features["quantization_matrix"]
    else:
        quantisation_matrix = QUANTISATION_MATRICES[
            (
                codec_features["wavelet_index"],
                codec_features["wavelet_index_ho"],
                codec_features["dwt_depth"],
                codec_features["dwt_depth_ho"],
            )
        ]

    dimensions_and_depths = compute_dimensions_and_depths(
        codec_features["video_parameters"],
        codec_features["picture_coding_mode"],
    )

    # Load the test pattern specifications
    try:
        (
            analysis_signal_bounds,
            synthesis_signal_bounds,
            analysis_test_patterns,
            synthesis_test_patterns,
        ) = bundle_get_static_filter_analysis(
            bundle_filename,
            codec_features["wavelet_index"],
            codec_features["wavelet_index_ho"],
            codec_features["dwt_depth"],
            codec_features["dwt_depth_ho"],
        )
    except KeyError:
        raise MissingStaticAnalysisError()

    all_analysis_pictures = []
    all_synthesis_pictures = []
    for component in ["Y", "C1"]:
        picture_bit_width = dimensions_and_depths[component].depth_bits
        picture_width = dimensions_and_depths[component].width
        picture_height = dimensions_and_depths[component].height

        # ...using optimised synthesis test patterns if available...
        try:
            synthesis_test_patterns = bundle_get_optimised_synthesis_test_patterns(
                bundle_filename,
                codec_features["wavelet_index"],
                codec_features["wavelet_index_ho"],
                codec_features["dwt_depth"],
                codec_features["dwt_depth_ho"],
                quantisation_matrix,
                picture_bit_width,
            )
        except KeyError:
            pass

        # Determine the maximum quantisation index which may be usefully used for
        # the codec configuration specified
        concrete_analysis_bounds, concrete_synthesis_bounds = evaluate_filter_bounds(
            codec_features["wavelet_index"],
            codec_features["wavelet_index_ho"],
            codec_features["dwt_depth"],
            codec_features["dwt_depth_ho"],
            analysis_signal_bounds,
            synthesis_signal_bounds,
            picture_bit_width,
        )
        max_quantisation_index = quantisation_index_bound(
            concrete_analysis_bounds,
            quantisation_matrix,
        )

        # Find the worst-case quantisation index for the synthesis test patterns
        _, synthesis_test_pattern_outputs = evaluate_test_pattern_outputs(
            codec_features["wavelet_index"],
            codec_features["wavelet_index_ho"],
            codec_features["dwt_depth"],
            codec_features["dwt_depth_ho"],
            picture_bit_width,
            quantisation_matrix,
            max_quantisation_index,
            analysis_test_patterns,
            synthesis_test_patterns,
        )

        # Generate packed test pictures
        analysis_pictures, synthesis_pictures = generate_test_pictures(
            picture_width,
            picture_height,
            picture_bit_width,
            analysis_test_patterns,
            synthesis_test_patterns,
            synthesis_test_pattern_outputs,
        )

        all_analysis_pictures.append(analysis_pictures)
        all_synthesis_pictures.append(synthesis_pictures)

    return (
        all_analysis_pictures[0],
        all_synthesis_pictures[0],
        all_analysis_pictures[1],
        all_synthesis_pictures[1],
    )



"""
Tests which verify that codecs correctly process pictures in a stream.
"""

import logging

from bitarray import bitarray

from itertools import cycle, count, islice

from copy import deepcopy

from enum import Enum

from vc2_data_tables import Profiles

from vc2_conformance.bitstream import Stream

from vc2_conformance.py2x_compat import zip

from vc2_conformance.pseudocode.slice_sizes import slice_bytes

from vc2_conformance.pseudocode.vc2_math import intlog2

from vc2_conformance.pseudocode.state import State

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.picture_generators import (
    moving_sprite,
    static_sprite,
    mid_gray,
    linear_ramps,
    white_noise,
)

from vc2_conformance.pseudocode.picture_encoding import filter_bit_shift

from vc2_conformance.encoder import make_sequence

from vc2_conformance.test_cases.decoder.common import (
    make_dummy_end_of_sequence,
    iter_slices_in_sequence,
)


def generate_filled_padding(
    padding_length_bits,
    filler=b"\x00",
    byte_align=0,
):
    """
    Generate a :py:class:`~bitarray.bitarray` ``padding_length_bits`` long
    containing repeated copies of ``filler``. If byte_align is non-zero, the
    generated bit array will have a sufficient number of 0s at the start to
    byte-align the first copy of ``filler`` assuming the padding starts at
    a bit address specified by ``byte_align``.

    Parameters
    ==========
    padding_length_bits : int
        Number of padding bits to generate.
    filler : bytes
        Filler bytes to be used.
    byte_align : int
        If given, byte-align the filler assuming that the padding starts from
        the provided bit-address.
    """
    byte_alignment_bits = (8 - (byte_align % 8)) % 8

    required_bytes = (padding_length_bits - byte_alignment_bits + 7) // 8
    filler_repeats = (required_bytes + len(filler) - 1) // len(filler)

    # Add repeated filler
    padding = bitarray()
    padding.frombytes(filler * filler_repeats)

    # Add byte-alignment bits
    if byte_alignment_bits:
        padding = bitarray("0" * byte_alignment_bits) + padding

    # Trim any excess
    return padding[:padding_length_bits]


def fill_hq_slice_padding(
    state,
    sx,
    sy,
    hq_slice,
    component,
    filler,
    byte_align=False,
    min_length=0,
):
    """
    Given a :py:class:`~vc2_conformance.bitstream.HQSlice`, mutate it in-place
    such that the slice size remains the same, (or becomes
    min_length*slice_size_scaler + 4 bytes, whichever is greater) but the
    specified component's bounded block padding data is set to repeated copies
    of ``filler``. All transform data will be forced to zeros.

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least ``slice_size_scaler``.
    sx, sy : int
        The slice coordinates. (Not used but present for consistency with
        :py:func:`fill_ld_slice_padding`.
    hq_slice : :py:class:`~vc2_conformance.bitstream.HQSlice`
        The slice to modify.
    component : "Y", "C1" or "C2"
        The component to stuff with filler data.
    filler : bytes
        A byte string to use as filler. Will be repeated as many times as
        necessary to fill the available space.
    byte_align : bool
        If True, insert the filler bytes on a byte-aligned boundary.
    min_length : int
        The minimum total length field for the modified slice
    """
    assert len(filler) > 0

    # Force picture content to zeros
    for c in ["y_transform", "c1_transform", "c2_transform"]:
        for i in range(len(hq_slice[c])):
            hq_slice[c][i] = 0

    # Make the designated slice take up the full size
    total_length = (
        hq_slice["slice_y_length"]
        + hq_slice["slice_c1_length"]
        + hq_slice["slice_c2_length"]
    )

    total_length = max(min_length, total_length)

    hq_slice["slice_y_length"] = 0
    hq_slice["slice_c1_length"] = 0
    hq_slice["slice_c2_length"] = 0
    hq_slice["slice_{}_length".format(component.lower())] = total_length

    # Work out how many padding bits we have to fill
    total_bytes = total_length * state["slice_size_scaler"]
    total_bits = total_bytes * 8
    bits_used_by_zeros = len(hq_slice["{}_transform".format(component.lower())])
    padding_bits = total_bits - bits_used_by_zeros

    # Generate the padding
    if padding_bits > 0:
        padding = generate_filled_padding(
            padding_bits,
            filler,
            # NB: All blocks in a HQ slice are byte aligned so all we need to
            # compensate for is the zeroed transform coeffs themselves
            bits_used_by_zeros if byte_align else 0,
        )
    else:
        # Zero or -ve padding (i.e. there is no space for padding)
        padding = bitarray()

    hq_slice["{}_block_padding".format(component.lower())] = padding


def fill_ld_slice_padding(
    state,
    sx,
    sy,
    ld_slice,
    component,
    filler,
    byte_align=False,
):
    """
    Given a :py:class:`~vc2_conformance.bitstream.LDSlice`, mutate it in-place
    such that the slice size remains the same, but the
    specified component's bounded block padding data is set to repeated copies
    of ``filler``. All transform data will be forced to zeros.

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least:

        * ``slices_x``
        * ``slices_y``
        * ``slice_bytes_numerator``
        * ``slice_bytes_denominator``
    sx, sy : int
        The slice coordinates
    ld_slice : :py:class:`~vc2_conformance.bitstream.LDSlice`
        The slice to modify.
    component : "Y", "C"
        The component to stuff with filler data.
    filler : bytes
        A byte string to use as filler. Will be repeated as many times as
        necessary to fill the available space.
    byte_align : bool
        If True, insert the filler bytes on a byte-aligned boundary.
    """
    assert len(filler) > 0

    # Force picture content to zeros
    for c in ["y_transform", "c_transform"]:
        for i in range(len(ld_slice[c])):
            ld_slice[c][i] = 0

    # Work out the slice's size
    slice_data_bits = 8 * slice_bytes(state, sx, sy)
    slice_data_bits -= 7
    length_field_bits = intlog2(slice_data_bits)
    slice_data_bits -= length_field_bits

    # Force the specified component to the full length
    if component == "Y":
        ld_slice["slice_y_length"] = slice_data_bits
    else:
        ld_slice["slice_y_length"] = 0

    # Work out the size of the padding bits
    bits_used_by_zeros = len(ld_slice["{}_transform".format(component.lower())])
    padding_bits = slice_data_bits - bits_used_by_zeros

    data_start_offset_bits = 7 + length_field_bits

    if padding_bits > 0:
        padding = generate_filled_padding(
            padding_bits,
            filler,
            data_start_offset_bits if byte_align else 0,
        )
    else:
        padding = bitarray()

    ld_slice["{}_block_padding".format(component.lower())] = padding


def generate_exp_golomb_with_ascending_lengths(sign=1):
    """
    Generate a series of integers whose signed exp-golmb encodings have
    sequentially growing lengths.

    Yields an infinite series of (length, int) pairs with the same sign as the
    ``sign`` argument.
    """
    yield (1, 0)

    for int_bits, exp_golomb_bits in zip(count(1), count(4, 2)):
        yield (
            exp_golomb_bits,
            ((1 << int_bits) - 1) * (1 if sign > 0 else -1),
        )


class UnsatisfiableBlockSizeError(ValueError):
    """
    Thrown when trying to generate values for a particular block fails.
    """


def generate_exp_golomb_with_length(num_values, required_length):
    """
    Given a target number of bits, produce ``num_values`` whose signed
    exp-Golomb encoding is ``required_length`` bits long. Raises a
    :py:exc:`UnsatisfiableBlockSizeError` if the required length cannot be
    achieved with ``num_values``.

    This function tries to keep the length and the magnitudes of the returned
    numbers as low as possible. Returned numbers will alternate in sign.

    Parameters
    ==========
    num_values : int
        The number of values to return.
    required_length : int
        The number of bits which the returned values must take up when encoded
        as signed exp-golomb codes.

    Returns
    =======
    values : [int, ...]
    """
    # The basis of the algorithm used by this function is as follows. We start
    # out with num_values zeros. One by one, we increment each value to
    # increase its length, e.g. for num_values == 4:
    #
    #     Step 0: [0,  0, 0,  0]    Step 5: [3, -1, 1, -1]
    #     Step 1: [1,  0, 0,  0]    Step 6: [3, -3, 1, -1]
    #     Step 2: [1, -1, 0,  0]    Step 7: [3, -3, 3, -1]
    #     Step 3: [1, -1, 1,  0]    Step 8: [3, -3, 3, -3]
    #     Step 4: [1, -1, 1, -1]    Step 9: [7, -3, 3, -3]
    #
    # The algorithm then terminates successfully when the total length reaches
    # the required amount, or fails if it exceeds it.
    #
    # For the first num_values steps, the number of bits required increases
    # from num_values to num_values + 3*num_values. For every subsequent step,
    # the number of bits increases by 2. Clearly, then, there are some values
    # which this algorithm will never be able to reach because they don't sit
    # on a multiple of 2 or 3 boundary.
    #
    # To work-around the multiple-of-2-or-3 problem, if the algoritm fails to
    # find a match, the whole process is repeated with the last one or two
    # values fixed at zero, offsetting the starting number of bits by 1 or 2,
    # potentially enabling a solution to be found.

    for num_values, required_length, extra_zeros in [
        (num_values, required_length, []),
        (num_values - 1, required_length - 1, [0]),
        (num_values - 2, required_length - 2, [0, 0]),
    ]:
        # Skip impossible lengths
        if num_values < 0 or required_length < 0:
            continue
        elif num_values == 0 and required_length != 0:
            continue

        # For each value in num_values, an iterator which generates increasingly
        # long (length, int) pairs.
        iterators = [
            iter(generate_exp_golomb_with_ascending_lengths(sign))
            for sign in islice(cycle([+1, -1]), num_values)
        ]

        lengths_and_values = [next(it) for it in iterators]

        iterator_to_increment = cycle(range(len(iterators)))

        while True:
            current_length = sum(length for length, value in lengths_and_values)
            if current_length == required_length:
                return [value for length, value in lengths_and_values] + extra_zeros
            elif current_length < required_length:
                i = next(iterator_to_increment)
                lengths_and_values[i] = next(iterators[i])
            else:  # if current_length > required_length:
                break

    raise UnsatisfiableBlockSizeError(
        "Cannot create a series of {} signed integers "
        "exactly {} bits long.".format(
            num_values,
            required_length,
        )
    )


class DanglingTransformValueType(Enum):
    """
    The four different ways a transform coefficient can dangle off the end of a
    bounded block.
    """

    zero_dangling = 0
    """
    A zero value (1 bit) is encoded entirely beyond the end of the bounded
    block.
    """

    sign_dangling = 1
    """
    The final bit (the sign bit) of a non-zero exp-golomb value is dangling
    beyond the end of the bounded block.
    """

    stop_and_sign_dangling = 2
    """
    The final two bits (the stop bit and sign bit) of a non-zero exp-golomb
    value are dangling beyond the end of the bounded block.
    """

    lsb_stop_and_sign_dangling = 3
    """
    The final three bits (the least significant bit, stop bit and sign bit) of
    a non-zero exp-golomb value are dangling beyond the end of the bounded
    block.
    """


def generate_dangling_transform_values(block_bits, num_values, dangle_type, magnitude):
    """
    Generates sets of values which can be encoded into a ``block_bits`` bit
    long bounded block where some value is entirely or partly beyond the end of
    the bounded block.

    Parameters
    ==========
    block_bits : int
        The number of bits in the bounded block.
    num_values : int
        The number of values to be returned.
    dangle_type : :py:class:`DanglingTransformValueType`
        The type of dangling value to produce.
    magnitude : int
        The magnitude of the dangling value, in bits (for two's compliment
        representation). Ignored when dangle_type is
        :py:attr:`DanglingTransformValueType.zero_dangling`. When
        :py:attr:`DanglingTransformValueType.lsb_stop_and_sign_dangling`, the
        two's complement representation will actually be magnitude+1 bits.

    Returns
    =======
    values : [value, ...]
        ``num_values`` transform coefficients which, when encoded in a bounded
        block ``block_bits`` long will contain a dangling value of the
        requested type.

    Raises
    ======
        Raises :py:exc:`UnsatisfiableBlockSizeError` if no series of
        ``num_values`` could be produced meeting these requirements. This
        should only occur for degenerate cases (e.g.  near-zero block_bits or
        num_values or large magnitude values).
    """
    # Degenerate case: Can't do anything with no values...
    if num_values == 0:
        raise UnsatisfiableBlockSizeError("Number of values is 0")

    # Pick a value to dangle over the end of the block and the number of bits
    # of the value to be kept *inside* the block
    exp_golomb_length = (2 * magnitude) + 2
    value, length = {
        DanglingTransformValueType.zero_dangling: (0, 0),
        DanglingTransformValueType.sign_dangling: (
            -((1 << magnitude) - 1),
            exp_golomb_length - 1,
        ),
        DanglingTransformValueType.stop_and_sign_dangling: (
            -((1 << magnitude) - 1),
            exp_golomb_length - 2,
        ),
        DanglingTransformValueType.lsb_stop_and_sign_dangling: (
            -(1 << magnitude),
            exp_golomb_length - 3,
        ),
    }[dangle_type]

    # Degenerate case: Can't fit the non-dangling part within the allowed space
    if length > block_bits:
        raise UnsatisfiableBlockSizeError("Block too small for non-dangling part.")

    # Put as many non-zero values in-front of our target value as possible such
    # that all but 'length' bits of the bounded block are filled.
    values_before = []
    # NB: Range below means we try at most num_values - 1 values, i.e. we
    # always leave one value to use as our dangling value
    for num_values_before in reversed(
        range(
            min(
                block_bits - length + 1,
                num_values,
            )
        )
    ):
        try:
            values_before = generate_exp_golomb_with_length(
                num_values_before,
                block_bits - length,
            )
            break
        except UnsatisfiableBlockSizeError:
            values_before = None

    # Degenrate case: we can't pad the leading space in the block to fill all
    # but length bits. Should only occur for ludicrously small blocks/lengths.
    if values_before is None:
        raise UnsatisfiableBlockSizeError("Can't align dangling value.")

    return values_before + [value] + [0] * (num_values - len(values_before) - 1)


def cut_off_value_at_end_of_hq_slice(
    state,
    sx,
    sy,
    hq_slice,
    component,
    dangle_type,
    magnitude,
    min_length=0,
):
    """
    Given a :py:class:`~vc2_conformance.bitstream.HQSlice`, mutate it in-place
    such that the slice size remains the same, (or becomes
    min_length*slice_size_scaler + 4 bytes, whichever is greater) but the
    specified component's bounded block contains a value whose last bits are
    off the end of the bounded block.

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least ``slice_size_scaler``.
    sx, sy : int
        The slice coordinates. (Not used but present for consistency with
        :py:func:`cut_off_value_at_end_of_hq_slice`.
    hq_slice : :py:class:`~vc2_conformance.bitstream.HQSlice`
        The slice to modify.
    component : "Y", "C1" or "C2"
        The component in which to place the dangling value.
    dangle_type : :py:class:`DanglingTransformValueType`
        The type of dangling value to produce.
    magnitude : int
        The magnitude of the dangling value, in bits. Ignored when dangle_type
        is :py:attr:`DanglingTransformValueType.lsb_stop_and_sign_dangling`.
    min_length : int
        The minimum total length field for the modified slice

    Raises
    ======
    UnsatisfiableBlockSizeError
        If unable to create the specified type of dangling values in the space
        and transform components available. Should only occur in degenerate
        cases.
    """
    # Make the designated slice have at least 16 bits (this should allow for
    # plenty of flexibility
    total_length = (
        hq_slice["slice_y_length"]
        + hq_slice["slice_c1_length"]
        + hq_slice["slice_c2_length"]
    )

    total_length = max(min_length, total_length)

    if total_length < 2:
        raise UnsatisfiableBlockSizeError("Slices too small")

    if component == "Y":
        hq_slice["slice_y_length"] = 2
        hq_slice["slice_c1_length"] = 0
        hq_slice["slice_c2_length"] = total_length - 2
    elif component == "C1":
        hq_slice["slice_y_length"] = 0
        hq_slice["slice_c1_length"] = 2
        hq_slice["slice_c2_length"] = total_length - 2
    elif component == "C2":
        hq_slice["slice_y_length"] = 0
        hq_slice["slice_c1_length"] = total_length - 2
        hq_slice["slice_c2_length"] = 2

    # Work out length of component to truncate
    block_bytes = 2 * state["slice_size_scaler"]
    block_bits = block_bytes * 8

    # Work out how many transform coeffs are in this component
    num_values = len(hq_slice["{}_transform".format(component.lower())])

    # Force picture content to zeros
    for c in ["y_transform", "c1_transform", "c2_transform"]:
        for i in range(len(hq_slice[c])):
            hq_slice[c][i] = 0

    # Fill requested component with dangling values
    values = generate_dangling_transform_values(
        block_bits,
        num_values,
        dangle_type,
        magnitude,
    )
    hq_slice["{}_transform".format(component.lower())] = values


def cut_off_value_at_end_of_ld_slice(
    state,
    sx,
    sy,
    ld_slice,
    component,
    dangle_type,
    magnitude,
):
    """
    Given a :py:class:`~vc2_conformance.bitstream.LDSlice`, mutate it in-place
    such that the slice size remains the same, but the specified component's
    bounded block contains a value whose last bits are off the end of the
    bounded block.

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least:

        * ``slices_x``
        * ``slices_y``
        * ``slice_bytes_numerator``
        * ``slice_bytes_denominator``
    sx, sy : int
        The slice coordinates. (Not used but present for consistency with
        :py:func:`cut_off_value_at_end_of_hq_slice`.
    ld_slice : :py:class:`~vc2_conformance.bitstream.LDSlice`
        The slice to modify.
    component : "Y" or "C"
        The component in which to place the dangling value.
    dangle_type : :py:class:`DanglingTransformValueType`
        The type of dangling value to produce.
    magnitude : int
        The magnitude of the dangling value, in bits. Ignored when dangle_type
        is :py:attr:`DanglingTransformValueType.lsb_stop_and_sign_dangling`.

    Raises
    ======
    UnsatisfiableBlockSizeError
        If unable to create the specified type of dangling values in the space
        and transform components available. Should only occur in degenerate
        cases.
    """
    # Make the designated slice have at least 16 bits (this should allow for
    # plenty of flexibility in fitting sensible magnitudes
    slice_data_bits = 8 * slice_bytes(state, sx, sy)
    slice_data_bits -= 7
    length_field_bits = intlog2(slice_data_bits)
    slice_data_bits -= length_field_bits

    block_bits = 16

    if slice_data_bits < block_bits:
        raise UnsatisfiableBlockSizeError("Slices too small")

    if component == "Y":
        ld_slice["slice_y_length"] = block_bits
    elif component == "C":
        ld_slice["slice_y_length"] = slice_data_bits - block_bits

    # Work out how many transform coeffs are in this component
    num_values = len(ld_slice["{}_transform".format(component.lower())])

    # Force picture content to zeros
    for c in ["y_transform", "c_transform"]:
        for i in range(len(ld_slice[c])):
            ld_slice[c][i] = 0

    # Fill requested component with dangling values
    values = generate_dangling_transform_values(
        block_bits,
        num_values,
        dangle_type,
        magnitude,
    )
    ld_slice["{}_transform".format(component.lower())] = values


@decoder_test_case_generator
def slice_padding_data(codec_features):
    """
    **Tests that padding bits in picture slices are ignored.**

    Picture slices (13.5.3) and (13.5.4) might contain padding bits beyond the
    end of the transform coefficients for each picture component. These test
    cases check that decoders correctly ignore these values. Padding values
    will be filled with the following:

    ``slice_padding_data[slice_?_all_zeros]``
        Padding bits are all zero.

    ``slice_padding_data[slice_?_all_ones]``
        Padding bits are all one.

    ``slice_padding_data[slice_?_alternating_1s_and_0s]``
        Padding bits alternate between one and zero, starting with one.

    ``slice_padding_data[slice_?_alternating_0s_and_1s]``
        Padding bits alternate between zero and one, starting with zero.

    ``slice_padding_data[slice_?_dummy_end_of_sequence]``
        Padding bits will contain bits which encode an end of sequence data
        unit (10.6).

    The above cases are repeated for the luma and color difference picture
    components as indicated by the value substituted for ``?`` in the test case
    names above. For low-delay pictures these will be ``Y`` (luma) and ``C``
    (interleaved color difference). For high quality pictures these will be
    ``Y`` (luma), ``C1`` (color difference 1) and ``C2`` (color difference 2).
    """
    # The values with which to fill padding data
    #
    # [(filler, byte_align, explanation), ...]
    filler_values = [
        (b"\x00", False, "all_zeros"),
        (b"\xFF", False, "all_ones"),
        (b"\xAA", False, "alternating_1s_and_0s"),
        (b"\x55", False, "alternating_0s_and_1s"),
        (make_dummy_end_of_sequence(), True, "dummy_end_of_sequence"),
    ]

    # The picture components expected
    if codec_features["profile"] == Profiles.high_quality:
        picture_components = ["Y", "C1", "C2"]
    elif codec_features["profile"] == Profiles.low_delay:
        picture_components = ["Y", "C"]

    # Generate single-frame mid-gray sequences with the specified padding data
    base_sequence = make_sequence(
        codec_features,
        # These pictures encode to all zeros which should give the highest
        # possible compression.
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
    )

    for filler, byte_align, explanation in filler_values:
        for component in picture_components:
            sequence = deepcopy(base_sequence)
            for (state, sx, sy, slice) in iter_slices_in_sequence(
                codec_features,
                sequence,
            ):
                if codec_features["profile"] == Profiles.high_quality:
                    # For lossless coding, extend the slice size to ensure some
                    # padding data is used
                    if codec_features["lossless"]:
                        min_length = (
                            slice["slice_y_length"]
                            + slice["slice_c1_length"]
                            + slice["slice_c2_length"]
                            + 8
                        )
                    else:
                        min_length = 0

                    fill_hq_slice_padding(
                        state,
                        sx,
                        sy,
                        slice,
                        component,
                        filler,
                        byte_align,
                        min_length,
                    )
                elif codec_features["profile"] == Profiles.low_delay:
                    fill_ld_slice_padding(
                        state,
                        sx,
                        sy,
                        slice,
                        component,
                        filler,
                        byte_align,
                    )

            yield TestCase(
                Stream(sequences=[sequence]),
                "{}_{}".format(
                    component,
                    explanation,
                ),
            )


@decoder_test_case_generator
def dangling_bounded_block_data(codec_features):
    """
    **Tests that transform values which lie beyond the end of a bounded block
    are read correctly.**

    Picture slices (13.5.3) and (13.5.4) contain transform values in bounded
    blocks (A.4.2). These test cases include bounded blocks in which some
    encoded values lie off the end of the block. Specifically, the following
    cases are tested:

    ``dangling_bounded_block_data[zero_dangling]``
        .. image:: /_static/user_guide/dangling_bounded_block_data_zero_dangling.svg

        A zero value (1 bit) is encoded entirely beyond the end of the bounded
        block.

    ``dangling_bounded_block_data[sign_dangling]``
        .. image:: /_static/user_guide/dangling_bounded_block_data_sign_dangling.svg

        The final bit (the sign bit) of a non-zero exp-golomb value is dangling
        beyond the end of the bounded block.

    ``dangling_bounded_block_data[stop_and_sign_dangling]``
        .. image:: /_static/user_guide/dangling_bounded_block_data_stop_and_sign_dangling.svg

        The final two bits (the stop bit and sign bit) of a non-zero exp-golomb
        value are dangling beyond the end of the bounded block.

    ``dangling_bounded_block_data[lsb_stop_and_sign_dangling]``
        .. image:: /_static/user_guide/dangling_bounded_block_data_lsb_stop_and_sign_dangling.svg

        The final three bits (the least significant bit, stop bit and sign bit)
        of a non-zero exp-golomb value are dangling beyond the end of the
        bounded block.

    .. note::

        The value and magnitudes of the dangling values are chosen depending on
        the wavelet transform in use and might differ from the illustrations
        above.
    """
    # The magnitude of the dangling value is chosen such that even if it ends
    # up being part of the DC component, the bit-shift used by some wavelets
    # won't make it disappear entirely.
    shift = filter_bit_shift(
        State(
            wavelet_index=codec_features["wavelet_index"],
            wavelet_index_ho=codec_features["wavelet_index_ho"],
        )
    )
    magnitude = (
        (codec_features["dwt_depth"] + codec_features["dwt_depth_ho"]) * shift
    ) + 1

    # The picture components expected
    if codec_features["profile"] == Profiles.high_quality:
        picture_components = ["Y", "C1", "C2"]
    elif codec_features["profile"] == Profiles.low_delay:
        picture_components = ["Y", "C"]

    # Generate single-frame mid-gray sequences
    base_sequence = make_sequence(
        codec_features,
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
    )

    # Replace with dangling values as required
    for dangle_type in DanglingTransformValueType:
        for component in picture_components:
            try:
                sequence = deepcopy(base_sequence)
                for (state, sx, sy, slice) in iter_slices_in_sequence(
                    codec_features,
                    sequence,
                ):
                    if codec_features["profile"] == Profiles.high_quality:
                        # For lossless coding, extend the slice size to ensure some
                        # data is used
                        if codec_features["lossless"]:
                            min_length = 2
                        else:
                            min_length = 0

                        cut_off_value_at_end_of_hq_slice(
                            state,
                            sx,
                            sy,
                            slice,
                            component,
                            dangle_type,
                            magnitude,
                            min_length,
                        )
                    elif codec_features["profile"] == Profiles.low_delay:
                        cut_off_value_at_end_of_ld_slice(
                            state,
                            sx,
                            sy,
                            slice,
                            component,
                            dangle_type,
                            magnitude,
                        )

                yield TestCase(
                    Stream(sequences=[sequence]),
                    "{}_{}".format(
                        dangle_type.name,
                        component,
                    ),
                )
            except UnsatisfiableBlockSizeError:
                logging.warning(
                    (
                        "Slices are too small to generate"
                        "dangling_bounded_block_data[%s_%s] test case."
                    ),
                    dangle_type.name,
                    component,
                )


@decoder_test_case_generator
def interlace_mode_and_pixel_aspect_ratio(codec_features):
    """
    **Tests that the interlacing mode and pixel aspect ratio is correctly
    decoded.**

    These tests require that the decoded pictures are observed using the
    intended display equipment for the decoder to ensure that the relevant
    display metadata is passed on.

    ``interlace_mode_and_pixel_aspect_ratio[static_sequence]``
        A single frame containing a stationary graphic at the top-left corner
        on a black background, as illustrated below.

        .. image:: /_static/user_guide/interlace_mode_and_pixel_aspect_ratio_static_sequence.svg

        If the field ordering (i.e. top field first flag, see (7.3) and (11.3))
        has been decoded correctly, the edges should be smooth. If the field
        order has been reversed the edges will appear jagged.

        If the pixel aspect ratio (see (11.4.7)) has been correctly decoded,
        the white triangle should be as wide as it is tall and the 'hole'
        should be circular.

    ``interlace_mode_and_pixel_aspect_ratio[moving_sequence]``
        A sequence of 10 frames containing a graphic moving from left to right
        along the top of the frame. In each successive frame, the graphic moves
        16 luma samples to the right (i.e. 8 samples every field, for
        interlaced formats).

        .. image:: /_static/user_guide/interlace_mode_and_pixel_aspect_ratio_moving_sequence.svg

        For progressive formats, the graphic should appear with smooth edges in
        each frame.

        For interlaced formats, the graphic should move smoothly when displayed
        on an interlaced monitor. If displayed as progressive frames (as in the
        illustration above), the pictures will appear to have ragged edges.
    """
    yield TestCase(
        Stream(
            sequences=[
                make_sequence(
                    codec_features,
                    static_sprite(
                        codec_features["video_parameters"],
                        codec_features["picture_coding_mode"],
                    ),
                )
            ]
        ),
        "static_sequence",
    )

    yield TestCase(
        Stream(
            sequences=[
                make_sequence(
                    codec_features,
                    moving_sprite(
                        codec_features["video_parameters"],
                        codec_features["picture_coding_mode"],
                    ),
                )
            ]
        ),
        "moving_sequence",
    )


@decoder_test_case_generator
def static_gray(codec_features):
    """
    **Tests that the decoder can decode a maximally compressible sequence.**

    This sequence contains an image in which every transform coefficient is
    zero. For most color specifications (11.4.10), this decodes to a mid-gray
    frame.

    This special case image is maximally compressible since no transform
    coefficients need to be explicitly coded in the bitstream. For lossless
    coding modes, this will also produce produce the smallest possible
    bitstream.
    """
    return Stream(
        sequences=[
            make_sequence(
                codec_features,
                mid_gray(
                    codec_features["video_parameters"],
                    codec_features["picture_coding_mode"],
                ),
            )
        ]
    )


@decoder_test_case_generator
def static_ramps(codec_features):
    """
    **Tests that decoder correctly reports color encoding information.**

    This test requires that the decoded pictures are observed using the
    intended display equipment for the decoder to ensure that the relevant
    color coding metadata is passed on.

    A static frame containing linear signal ramps for white and primary
    red, green and blue (in that order, from top-to-bottom) as illustrated
    below:

    .. image:: /_static/user_guide/static_ramps.png

    The color bands must be in the correct order (white, red, green, blue from
    top to bottom). If not, the color components might have been ordered
    incorrectly.

    The red, green and blue colors should correspond to the red, green and blue
    primaries for the color specification (11.4.10.2).

    .. note::

        When D-Cinema primaries are specified (preset color primaries index 3),
        red, green and blue are replaced with CIE X, Y and Z respectively. Note
        that these might not represent physically realisable colors.

    The left-most pixels in each band are notionally video black and the
    right-most pixels video white, red, green and blue (respectively). That is,
    oversaturated signals (e.g. 'super-blacks' and 'super-white') are not
    included.

    .. note::

        For lossy codecs, the decoded signal values might vary due to coding
        artefacts.

    The value ramps in the test picture are linear, meaning that the (linear)
    pixel values increase at a constant rate from left (black) to right
    (saturated white/red/green/blue). Due to the non-linear response of human
    vision, this will produce a non-linear brightness ramp which appears to
    quickly saturate. Further, when a non-linear transfer function is specified
    (11.4.10.4) the raw decoded picture values will not be linearly spaced.

    .. note::

        When the D-Cinema transfer function is specified (preset transfer
        function index 3), the saturated signals do not correspond to a
        non-linear signal value of 1.0 but instead approximately 0.97. This is
        because the D-Cinema transfer function allocates part of its nominal
        output range to over-saturated signals.
    """
    return Stream(
        sequences=[
            make_sequence(
                codec_features,
                linear_ramps(
                    codec_features["video_parameters"],
                    codec_features["picture_coding_mode"],
                ),
            )
        ]
    )


@decoder_test_case_generator
def static_noise(codec_features):
    """
    **Tests that decoder correctly decodes a noise plate.**

    A static frame containing pseudo-random uniform noise as illustrated below:

    .. image:: /_static/user_guide/noise.png
    """
    return Stream(
        sequences=[
            make_sequence(
                codec_features,
                white_noise(
                    codec_features["video_parameters"],
                    codec_features["picture_coding_mode"],
                ),
            )
        ]
    )
